import math
from collections import deque
from datetime import timedelta
from typing import cast
from unittest import TestCase

from spacepackets.ecss import PusTelecommand
from tmtccmd.tc import WaitEntry, QueueEntryHelper

# Required for eval calls
# noinspection PyUnresolvedReferences
from tmtccmd.tc import LogQueueEntry, RawTcEntry
from tmtccmd.tc.queue import QueueWrapper, QueueHelper


class TestTcQueue(TestCase):
    def test_queue(self):
        queue_wrapper = QueueWrapper(info=None, queue=deque())
        self.assertEqual(queue_wrapper.queue, deque())
        queue_helper = QueueHelper(queue_wrapper)
        queue_helper.add_wait(timedelta(seconds=2))
        self.assertEqual(len(queue_wrapper.queue), 1)
        wait_entry = cast(WaitEntry, queue_wrapper.queue.pop())
        self.assertTrue(wait_entry)
        self.assertFalse(wait_entry.is_tc())
        self.assertEqual(wait_entry.wait_time.total_seconds(), 2.0)
        self.assertEqual(len(queue_wrapper.queue), 0)
        pus_cmd = PusTelecommand(service=17, subservice=1)
        queue_helper.add_pus_tc(pus_cmd)
        self.assertEqual(len(queue_wrapper.queue), 1)
        queue_helper.add_log_cmd("Test String")
        queue_helper.add_raw_tc(bytes([0, 1, 2]))
        queue_helper.add_ccsds_tc(pus_cmd.to_space_packet())
        queue_helper.add_packet_delay(timedelta(seconds=3.0))
        print(queue_wrapper.queue)
        self.assertEqual(len(queue_wrapper.queue), 5)

        pus_entry = queue_wrapper.queue.popleft()
        self.assertTrue(pus_entry.is_tc())
        cast_wrapper = QueueEntryHelper(pus_entry)
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, pus_cmd)
        self.assertTrue(pus_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_wait_entry()
        log_entry = queue_wrapper.queue.popleft()
        self.assertFalse(log_entry.is_tc())
        cast_wrapper.base = log_entry
        log_entry = cast_wrapper.to_log_entry()
        self.assertTrue(log_entry)
        with self.assertRaises(TypeError):
            cast_wrapper.to_raw_tc_entry()
        self.assertEqual(log_entry.log_str, "Test String")
        test_obj = eval(f"{log_entry!r}")
        self.assertEqual(test_obj.log_str, log_entry.log_str)

        raw_entry = queue_wrapper.queue.popleft()
        self.assertTrue(raw_entry.is_tc())
        cast_wrapper.base = raw_entry
        raw_entry = cast_wrapper.to_raw_tc_entry()
        self.assertTrue(raw_entry)
        self.assertEqual(raw_entry.tc, bytes([0, 1, 2]))
        test_obj = eval(f"{raw_entry!r}")
        self.assertEqual(raw_entry.tc, test_obj.tc)

        space_packet_entry = queue_wrapper.queue.popleft()
        self.assertTrue(space_packet_entry.is_tc())
        cast_wrapper.base = space_packet_entry
        space_packet_entry = cast_wrapper.to_space_packet_entry()
        self.assertTrue(space_packet_entry)
        self.assertTrue(space_packet_entry.space_packet, pus_cmd.to_space_packet())

        packet_delay = queue_wrapper.queue.pop()
        self.assertFalse(packet_delay.is_tc())
        cast_wrapper.base = packet_delay
        packet_delay = cast_wrapper.to_packet_delay_entry()
        self.assertEqual(packet_delay.delay_time.total_seconds(), 3.0)



"""Argument parser module"""
import argparse
import sys
from typing import Optional, List
from dataclasses import dataclass

from tmtccmd.config.prompt import prompt_op_code, prompt_service
from tmtccmd.logging import get_console_logger

from .defs import CoreModeStrings, CoreModeList, CoreComInterfaces
from .hook import TmTcCfgHookBase


LOGGER = get_console_logger()


def get_default_descript_txt() -> str:
    from tmtccmd.utility.conf_util import AnsiColors

    return (
        f"{AnsiColors.GREEN}TMTC Client Command Line Interface\n"
        f"{AnsiColors.RESET}This application provides generic components to execute "
        f"TMTC commanding\n"
    )


def create_default_args_parser(
    descript_txt: Optional[str] = None,
) -> argparse.ArgumentParser:
    if descript_txt is None:
        descript_txt = get_default_descript_txt()
    return argparse.ArgumentParser(
        description=descript_txt, formatter_class=argparse.RawTextHelpFormatter
    )


@dataclass
class DefProcedureParams:
    service: str
    op_code: str


@dataclass
class TcParams:
    delay: float = 0.0
    apid: int = 0


@dataclass
class BackendParams:
    mode: str = ""
    com_if_id: str = ""
    listener: bool = False
    interactive: bool = False


@dataclass
class AppParams:
    use_gui: bool = False
    reduced_printout: bool = False
    use_ansi_colors: bool = True


class SetupParams:
    def __init__(
        self,
        def_proc_args: Optional[DefProcedureParams] = None,
        tc_params: TcParams = TcParams(),
        backend_params: BackendParams = BackendParams(),
        app_params: AppParams = AppParams(),
    ):
        self.def_proc_args = def_proc_args
        self.tc_params = tc_params
        self.backend_params = backend_params
        self.app_params = app_params

    @property
    def apid(self):
        return self.tc_params.apid

    @apid.setter
    def apid(self, apid):
        self.tc_params.apid = apid

    @property
    def use_gui(self):
        return self.app_params.use_gui

    @use_gui.setter
    def use_gui(self, use_gui):
        self.app_params.use_gui = use_gui

    @property
    def mode(self):
        return self.backend_params.mode

    @mode.setter
    def mode(self, mode: str):
        self.backend_params.mode = mode

    @property
    def com_if_id(self):
        return self.backend_params.com_if_id

    @com_if_id.setter
    def com_if_id(self, com_if_id):
        self.backend_params.com_if_id = com_if_id


def add_default_tmtccmd_args(parser: argparse.ArgumentParser):
    add_default_mode_arguments(parser)
    add_default_com_if_arguments(parser)
    add_generic_arguments(parser)
    add_cfdp_parser(parser)

    add_ethernet_arguments(parser)


def parse_default_tmtccmd_input_arguments(
    parser: argparse.ArgumentParser,
    print_known_args: bool = False,
    print_unknown_args: bool = False,
) -> (argparse.Namespace, List[str]):
    """Parses all input arguments
    :return: Input arguments contained in a special namespace and accessable by args.<variable>
    """

    if len(sys.argv) == 1:
        print("No input arguments specified. Run with -h to get list of arguments")

    args, unknown = parser.parse_known_args()

    if print_known_args:
        LOGGER.info("Printing known arguments:")
        for argument in vars(args):
            LOGGER.debug(argument + ": " + str(getattr(args, argument)))
    if print_unknown_args:
        LOGGER.info("Printing unknown arguments:")
        for argument in unknown:
            LOGGER.info(argument)

    if len(unknown) > 0:
        print(f"Unknown arguments detected: {unknown}")
    return args, unknown


def add_cfdp_parser(arg_parser: argparse.ArgumentParser):
    subparsers = arg_parser.add_subparsers(
        title="CFDP",
        description="CCSDS File Delivery Protocol commands",
        help="CCDSDS File Delivery Commands",
        dest="cfdp",
    )
    cfdp = subparsers.add_parser("cfdp")
    cfdp.add_argument("-p", "--proxy")
    cfdp.add_argument(
        "-f", "--file", dest="cfdp_file", help="CFDP target file", default=None
    )
    cfdp.add_argument(
        "-d",
        "--dest",
        dest="cfdp_dest",
        help="CFDP file destination path",
        default=None,
    )


def add_generic_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument(
        "-g", "--gui", help="Use GUI mode", action="store_true", default=False
    )
    arg_parser.add_argument(
        "-s",
        "--service",
        help="Procedure service code which is passed to the TC handler objects",
        default=None,
    )
    arg_parser.add_argument(
        "-o",
        "--op_code",
        help="Procedcure operation code, which is passed to the TC packer functions",
        default=None,
    )
    arg_parser.add_argument(
        "-l",
        "--listener",
        help="The backend will be configured to go into listener mode after "
        "finishing the first queue",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-i",
        "--interactive",
        help="Enables interactive or multi-queue mode, where the backend will be configured "
        "to handle multiple queues",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-d",
        "--delay",
        type=float,
        help="Default inter-packet delay. Default: 3 seconds for one queue mode, "
        "0 for interactive mode",
        default=None,
    )


def add_default_mode_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config import CoreModeList, CoreModeStrings

    help_text = f"Core Modes. Default: {CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]}\n"
    one_q = (
        f"{CoreModeList.ONE_QUEUE_MODE} or "
        f"{CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]}: "
        f"One Queue Command Mode\n"
    )
    listener_help = (
        f"{CoreModeList.LISTENER_MODE} or {CoreModeStrings[CoreModeList.LISTENER_MODE]}: "
        f"Listener Mode\n"
    )
    multi_q = (
        f"{CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE} or "
        f"{CoreModeStrings[CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE]}: "
        f"Multi Queue and Interactive Command Mode\n"
    )
    help_text += one_q + listener_help + multi_q
    arg_parser.add_argument(
        "-m",
        "--mode",
        type=str,
        help=help_text,
        default=CoreModeStrings[CoreModeList.ONE_QUEUE_MODE],
    )


def add_default_com_if_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config import CORE_COM_IF_DICT, CoreComInterfaces

    help_text = (
        "Core Communication Interface. If this is not specified, the commander core\n"
        "will try to extract it from the JSON or prompt it from the user\n"
    )
    dummy_line = f"{CORE_COM_IF_DICT[CoreComInterfaces.DUMMY.value]}: Dummy Interface\n"
    udp_line = f"{CORE_COM_IF_DICT[CoreComInterfaces.UDP.value]}: " f"UDP client\n"
    ser_dle_line = (
        f"{CORE_COM_IF_DICT[CoreComInterfaces.SERIAL_DLE.value]}: "
        f"Serial with DLE transport layer\n"
    )
    ser_fixed_line = (
        f"{CORE_COM_IF_DICT[CoreComInterfaces.SERIAL_FIXED_FRAME.value]}: "
        f"Serial with fixed frames\n"
    )
    ser_qemu_line = (
        f"{CORE_COM_IF_DICT[CoreComInterfaces.SERIAL_QEMU.value]}: "
        f"QEMU serial interface\n"
    )
    help_text += dummy_line + ser_dle_line + udp_line + ser_fixed_line + ser_qemu_line
    arg_parser.add_argument(
        "-c",
        "--com_if",
        type=str,
        help=help_text,
        default=CoreComInterfaces.UNSPECIFIED.value,
    )


def add_ethernet_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument("--h-ip", help="Host (Computer) IP. Default:''", default="")
    arg_parser.add_argument(
        "--t-ip", help="Target IP. Default: Localhost 127.0.0.1", default="127.0.0.1"
    )


def args_to_params(
    pargs: argparse.Namespace,
    params: SetupParams,
    hook_obj: TmTcCfgHookBase,
    use_prompts: bool,
):
    """If some arguments are unspecified, they are set here with (variable) default values.

    :param pargs: Parsed arguments from calling parse method
    :param params: Setup parameter object which will be set by this function
    :param hook_obj:
    :param use_prompts: Specify whether terminal prompts are allowed to retrieve unspecified
        arguments. For something like a GUI, it might make sense to disable this
    :return: None
    """
    from tmtccmd.com_if.utils import determine_com_if

    if pargs.gui is None:
        params.app_params.use_gui = False
    else:
        params.app_params.use_gui = pargs.gui
    if pargs.com_if is None or pargs.com_if == CoreComInterfaces.UNSPECIFIED.value:
        params.com_if_id = determine_com_if(
            hook_obj.get_com_if_dict(), hook_obj.json_cfg_path, use_prompts
        )
    else:
        # TODO: Check whether COM IF is valid?
        params.com_if_id = pargs.com_if
    if pargs.mode is None:
        params.mode = CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]
    else:
        params.mode = pargs.mode
    tmtc_defs = hook_obj.get_tmtc_definitions()
    params.def_proc_args = DefProcedureParams("0", "0")
    if tmtc_defs is None:
        LOGGER.warning("Invalid Service to Op-Code dictionary detected")
    else:
        if pargs.service is None:
            if pargs.mode == CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]:
                if use_prompts:
                    print("No service argument (-s) specified, prompting from user")
                    # Try to get the service list from the hook base and prompt service from user
                    params.def_proc_args.service = prompt_service(tmtc_defs)
        else:
            params.def_proc_args.service = pargs.service
        if pargs.op_code is None:
            current_service = params.def_proc_args.service
            if use_prompts:
                params.def_proc_args.op_code = prompt_op_code(
                    tmtc_defs, current_service
                )
        else:
            params.def_proc_args.op_code = pargs.op_code
    if pargs.delay is None:
        if params.backend_params.mode == CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]:
            params.tc_params.delay = 3.0
        else:
            params.tc_params.delay = 0.0
    else:
        params.tc_params.delay = pargs.delay
    if pargs.listener is None:
        params.backend_params.listener = False
    else:
        params.backend_params.listener = pargs.listener


class ArgParserWrapper:
    def __init__(
        self,
        hook_obj: TmTcCfgHookBase,
        parser: Optional[argparse.ArgumentParser] = None,
        descript_txt: Optional[str] = None,
    ):
        if parser is None:
            self.args_parser = create_default_args_parser(descript_txt)
            add_default_tmtccmd_args(self.args_parser)
        else:
            self.args_parser = parser
        self.print_known_args = False
        self.print_unknown_args = False
        self.hook_obj = hook_obj
        self.unknown_args = [""]
        self.args_raw = None
        self._parse_was_called = False

    def add_default_tmtccmd_args(self):
        add_default_tmtccmd_args(self.args_parser)

    def parse(self):
        """Parse all CLI arguments with the given argument parser"""
        if not self._parse_was_called:
            self.args_raw, self.unknown_args = parse_default_tmtccmd_input_arguments(
                self.args_parser,
                print_known_args=self.print_known_args,
                print_unknown_args=self.print_unknown_args,
            )
        self._parse_was_called = True

    @property
    def use_gui(self):
        """This only yields valid values if :py:meth:`parse` was called once"""
        if not self._parse_was_called:
            return False
        return self.args_raw.gui

    def set_params(self, params: SetupParams):
        """Set up the parameter object from the parsed arguments. This call auto-determines whether
        prompts should be used depending on whether the GUI flag was passed or not.

        :raise Value Error: Parse function call missing
        """
        if not self._parse_was_called:
            raise ValueError("Call the parse function first")
        if self.args_raw.gui:
            self.set_params_without_prompts(params)
        else:
            self.set_params_with_prompts(params)

    def set_params_without_prompts(self, params: SetupParams):
        if not self._parse_was_called:
            raise ValueError("Call the parse function first")
        args_to_params(
            pargs=self.args_raw,
            params=params,
            hook_obj=self.hook_obj,
            use_prompts=False,
        )

    def set_params_with_prompts(self, params: SetupParams):
        if not self._parse_was_called:
            raise ValueError("Call the parse function first")
        try:
            args_to_params(
                pargs=self.args_raw,
                params=params,
                hook_obj=self.hook_obj,
                use_prompts=True,
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt(
                "Keyboard interrupt while converting CLI args to application parameters"
            )



from __future__ import annotations
from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss import PusVersion, PusTelemetry
from spacepackets.ecss.pus_17_test import Service17Tm
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase

from tmtccmd.pus.pus_17_test import Subservices


class Service17TmExtended(PusTmBase, PusTmInfoBase, Service17Tm):
    def __init__(
        self,
        subservice: int,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        Service17Tm.__init__(
            self,
            subservice=subservice,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=self.pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=self.pus_tm)
        self.__set_internal_fields()

    @classmethod
    def __empty(cls) -> Service17TmExtended:
        return cls(subservice=0)

    def __set_internal_fields(self):
        if self.subservice == Subservices.TM_REPLY:
            self.set_packet_info("Ping Reply")

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytes,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service17TmExtended:
        service_17_tm = cls.__empty()
        service_17_tm.pus_tm = PusTelemetry.unpack(raw_telemetry=raw_telemetry)
        service_17_tm.__set_internal_fields()
        return service_17_tm



from abc import abstractmethod, ABC
from pathlib import Path


class ProvidesSeqCount(ABC):
    @abstractmethod
    def next_seq_count(self) -> int:
        raise NotImplementedError(
            "Please use a concrete class implementing this method"
        )


class FileSeqCountProvider(ProvidesSeqCount):
    def __init__(self, file_name: Path = Path("seqcnt.txt")):
        self.file_name = file_name
        if not self.file_name.exists():
            self.create_new()

    def create_new(self):
        with open(self.file_name, "w") as file:
            file.write("0\n")

    def current(self) -> int:
        if not self.file_name.exists():
            raise FileNotFoundError(f"{self.file_name} file does not exist")
        with open(self.file_name) as file:
            return self.check_count(file.readline())

    def __next__(self):
        if not self.file_name.exists():
            raise FileNotFoundError(f"{self.file_name} file does not exist")
        with open(self.file_name, "r+") as file:
            curr_seq_cnt = self.increment_with_rollover(
                self.check_count(file.readline())
            )
            file.seek(0)
            file.write(f"{curr_seq_cnt}\n")
            return curr_seq_cnt

    @staticmethod
    def check_count(line: str) -> int:
        line = line.rstrip()
        if not line.isdigit():
            raise ValueError("Sequence count file content is invalid")
        curr_seq_cnt = int(line)
        if curr_seq_cnt < 0 or curr_seq_cnt > pow(2, 14) - 1:
            raise ValueError("Sequence count in file has invalid value")
        return curr_seq_cnt

    def next_seq_count(self) -> int:
        return self.__next__()

    @staticmethod
    def increment_with_rollover(seq_cnt: int) -> int:
        """CCSDS Sequence count has maximum size of 14 bit. Rollover after that size by default"""
        if seq_cnt >= pow(2, 14) - 1:
            return 0
        else:
            return seq_cnt + 1



"""Contains core methods called by entry point files to setup and start a tmtccmd application"""
import sys
import os
from datetime import timedelta
from typing import Union, cast, Optional

from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend
from tmtccmd.core.base import FrontendBase
from tmtccmd.tm.ccsds_tm_listener import CcsdsTmListener
from tmtccmd.config import TmTcCfgHookBase, backend_mode_conversion, SetupWrapper
from tmtccmd.core.ccsds_backend import BackendBase
from tmtccmd.tm import TmTypes, TmHandlerBase, CcsdsTmHandler
from tmtccmd.core.globals_manager import update_global
from tmtccmd.logging import get_console_logger
from tmtccmd.config.globals import set_default_globals_pre_args_parsing
from tmtccmd.core import ModeWrapper
from tmtccmd.tc import DefaultProcedureInfo
from tmtccmd.tc.handler import TcHandlerBase

VERSION_MAJOR = 3
VERSION_MINOR = 0
VERSION_REVISION = 0

# I think this needs to be in string representation to be parsed so we can't
# use a formatted string here.
__version__ = "3.0.0rc1"


LOGGER = get_console_logger()

__SETUP_WAS_CALLED = False
__SETUP_FOR_GUI = False


def version() -> str:
    return __version__


def setup(setup_args: SetupWrapper):
    """This function needs to be called first before running the TMTC commander core. The setup
    arguments encapsulate all required arguments for the TMTC commander.

    :param setup_args:     Setup arguments
    """
    global __SETUP_WAS_CALLED, __SETUP_FOR_GUI

    if os.name == "nt":
        import colorama

        colorama.init()
    if setup_args.params.use_gui:
        set_default_globals_pre_args_parsing(setup_args.params.apid)
    if not setup_args.params.use_gui:
        __handle_cli_args_and_globals(setup_args)
    __SETUP_FOR_GUI = setup_args.params.use_gui
    __SETUP_WAS_CALLED = True


def start(
    tmtc_backend: BackendBase,
    hook_obj: TmTcCfgHookBase,
    tmtc_frontend: Optional[FrontendBase] = None,
    app_name: str = "TMTC Commander",
):
    """This is the primary function to run the TMTC commander. Users should call this function to
    start the TMTC commander. Please note that :py:func:`setup` needs to be
    called before this function.  You also need to build a TMTC backend
    instance and pass it to this call. You can use :py:func:`create_default_tmtc_backend`
    to create a generic backend.

    :param tmtc_backend:        Custom backend can be passed here. Otherwise, a default backend
                                will be created
    :param hook_obj:
    :param tmtc_frontend:       Custom frontend can be passed here. Otherwise, a default frontend
                                will be created
    :param app_name:            Name of application. Will be displayed in GUI
    :raises RunTimeError:  if :py:func:`setup` was not called before
    :return:
    """
    global __SETUP_WAS_CALLED, __SETUP_FOR_GUI
    if not __SETUP_WAS_CALLED:
        LOGGER.warning("setup_tmtccmd was not called first. Call it first")
        sys.exit(1)
    if __SETUP_FOR_GUI:
        __start_tmtc_commander_qt_gui(
            tmtc_frontend=tmtc_frontend,
            hook_obj=hook_obj,
            tmtc_backend=tmtc_backend,
            app_name=app_name,
        )
    else:
        __start_tmtc_commander_cli(tmtc_backend=tmtc_backend)


def init_printout(use_gui: bool):
    if use_gui:
        print(f"-- tmtccmd v{version()} GUI Mode --")
    else:
        print(f"-- tmtccmd v{version()} CLI Mode --")


# TODO: Remove globals altogether
def __handle_cli_args_and_globals(setup_args: SetupWrapper):
    set_default_globals_pre_args_parsing(setup_args.params.apid)


def __start_tmtc_commander_cli(tmtc_backend: BackendBase):
    tmtc_backend.open_com_if()


def __start_tmtc_commander_qt_gui(
    tmtc_backend: BackendBase,
    hook_obj: TmTcCfgHookBase,
    tmtc_frontend: Union[None, FrontendBase] = None,
    app_name: str = "TMTC Commander",
):
    global __SETUP_WAS_CALLED
    try:
        from PyQt5.QtWidgets import QApplication

        if not __SETUP_WAS_CALLED:
            LOGGER.warning("setup_tmtccmd was not called first. Call it first")
            sys.exit(1)
        app = QApplication([app_name])
        if tmtc_frontend is None:
            from tmtccmd.gui import TmTcFrontend
            from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend

            tmtc_frontend = TmTcFrontend(
                hook_obj=hook_obj,
                tmtc_backend=cast(CcsdsTmtcBackend, tmtc_backend),
                app_name=app_name,
            )
        tmtc_frontend.start(app)
    except ImportError as e:
        LOGGER.exception(e)
        sys.exit(1)


def create_default_tmtc_backend(
    setup_wrapper: SetupWrapper, tm_handler: TmHandlerBase, tc_handler: TcHandlerBase
) -> BackendBase:
    """Creates a default TMTC backend instance which can be passed to the tmtccmd runner

    :param tc_handler:
    :param setup_wrapper:
    :param tm_handler:
    :return:
    """
    global __SETUP_WAS_CALLED

    from typing import cast

    if not __SETUP_WAS_CALLED:
        LOGGER.warning("setup_tmtccmd was not called first. Call it first")
        sys.exit(1)
    if tm_handler is None:
        LOGGER.warning(
            "No TM Handler specified! Make sure to specify at least one TM handler"
        )
        sys.exit(1)
    else:
        if tm_handler.get_type() == TmTypes.CCSDS_SPACE_PACKETS:
            tm_handler = cast(CcsdsTmHandler, tm_handler)
    com_if = setup_wrapper.hook_obj.assign_communication_interface(
        com_if_key=setup_wrapper.params.com_if_id
    )
    tm_listener = CcsdsTmListener(tm_handler)
    mode_wrapper = ModeWrapper()
    backend_mode_conversion(setup_wrapper.params.mode, mode_wrapper)
    # The global variables are set by the argument parser.
    tmtc_backend = CcsdsTmtcBackend(
        com_if=com_if,
        tm_listener=tm_listener,
        tc_handler=tc_handler,
        tc_mode=mode_wrapper.tc_mode,
        tm_mode=mode_wrapper.tm_mode,
    )
    tmtc_backend.inter_cmd_delay = timedelta(
        seconds=setup_wrapper.params.tc_params.delay
    )
    tmtc_backend.current_procedure = DefaultProcedureInfo(
        setup_wrapper.params.def_proc_args.service,
        setup_wrapper.params.def_proc_args.op_code,
    )
    return tmtc_backend



from __future__ import annotations
from datetime import timedelta
from enum import Enum
from typing import Optional, Deque, cast, Any, Type

from spacepackets.ccsds import SpacePacket
from spacepackets.ecss import PusTelecommand
from tmtccmd.tc.procedure import TcProcedureBase


class TcQueueEntryType(Enum):
    PUS_TC = "pus-tc"
    CCSDS_TC = "ccsds-tc"
    RAW_TC = "raw-tc"
    CUSTOM = "custom"
    LOG = "log"
    WAIT = "wait"
    PACKET_DELAY = "set-delay"


class TcQueueEntryBase:
    """Generic TC queue entry abstraction. This allows filling the TC queue with custom objects"""

    def __init__(self, etype: TcQueueEntryType):
        self.etype = etype

    def is_tc(self) -> bool:
        """Check whether concrete object is an actual telecommand"""
        if (
            self.etype == TcQueueEntryType.PUS_TC
            or self.etype == TcQueueEntryType.RAW_TC
            or self.etype == TcQueueEntryType.CCSDS_TC
        ):
            return True
        return False


QueueDequeT = Deque[TcQueueEntryBase]


class PusTcEntry(TcQueueEntryBase):
    def __init__(self, pus_tc: PusTelecommand):
        super().__init__(TcQueueEntryType.PUS_TC)
        self.pus_tc = pus_tc

    def __repr__(self):
        return f"{self.__class__.__name__}({self.pus_tc!r})"


class SpacePacketEntry(TcQueueEntryBase):
    def __init__(self, space_packet: SpacePacket):
        super().__init__(TcQueueEntryType.CCSDS_TC)
        self.space_packet = space_packet

    def __repr__(self):
        return f"{self.__class__.__name__}({self.space_packet!r})"


class LogQueueEntry(TcQueueEntryBase):
    def __init__(self, log_str: str):
        super().__init__(TcQueueEntryType.LOG)
        self.log_str = log_str

    def __repr__(self):
        return f"{self.__class__.__name__}({self.log_str!r})"


class RawTcEntry(TcQueueEntryBase):
    def __init__(self, tc: bytes):
        super().__init__(TcQueueEntryType.RAW_TC)
        self.tc = tc

    def __repr__(self):
        return f"{self.__class__.__name__}({self.tc!r})"


class WaitEntry(TcQueueEntryBase):
    def __init__(self, wait_time: timedelta):
        super().__init__(TcQueueEntryType.WAIT)
        self.wait_time = wait_time

    @classmethod
    def from_millis(cls, millis: int) -> WaitEntry:
        return cls(timedelta(milliseconds=millis))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.wait_time!r})"


class PacketDelayEntry(TcQueueEntryBase):
    def __init__(self, delay_time: timedelta):
        super().__init__(TcQueueEntryType.PACKET_DELAY)
        self.delay_time = delay_time

    @classmethod
    def from_millis(cls, millis: int) -> PacketDelayEntry:
        return cls(timedelta(milliseconds=millis))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.delay_time!r})"


class QueueEntryHelper:
    def __init__(self, base: TcQueueEntryBase):
        self.base = base

    @property
    def is_tc(self) -> bool:
        return self.base.is_tc()

    @property
    def entry_type(self) -> TcQueueEntryType:
        return self.base.etype

    def __cast_internally(
        self,
        obj_type: Type[TcQueueEntryBase],
        obj: TcQueueEntryBase,
        expected_type: TcQueueEntryType,
    ) -> Any:
        if obj.etype != expected_type:
            raise TypeError(f"Invalid object {obj} for type {self.base.etype}")
        return cast(obj_type, obj)

    def to_log_entry(self) -> LogQueueEntry:
        return self.__cast_internally(LogQueueEntry, self.base, TcQueueEntryType.LOG)

    def to_pus_tc_entry(self) -> PusTcEntry:
        return self.__cast_internally(PusTcEntry, self.base, TcQueueEntryType.PUS_TC)

    def to_raw_tc_entry(self) -> RawTcEntry:
        return self.__cast_internally(RawTcEntry, self.base, TcQueueEntryType.RAW_TC)

    def to_wait_entry(self) -> WaitEntry:
        return self.__cast_internally(WaitEntry, self.base, TcQueueEntryType.WAIT)

    def to_packet_delay_entry(self) -> PacketDelayEntry:
        return self.__cast_internally(
            PacketDelayEntry, self.base, TcQueueEntryType.PACKET_DELAY
        )

    def to_space_packet_entry(self) -> SpacePacketEntry:
        return self.__cast_internally(
            SpacePacketEntry, self.base, TcQueueEntryType.CCSDS_TC
        )


class QueueWrapper:
    def __init__(
        self,
        info: Optional[TcProcedureBase],
        queue: Optional[QueueDequeT],
        inter_cmd_delay: timedelta = timedelta(milliseconds=0),
    ):
        self.info = info
        self.queue = queue
        self.inter_cmd_delay = inter_cmd_delay

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(info={self.info!r}, queue={self.queue!r}, "
            f"inter_cmd_delay={self.inter_cmd_delay!r})"
        )


class QueueHelper:
    def __init__(self, queue_wrapper: QueueWrapper):
        self.queue_wrapper = queue_wrapper

    def __repr__(self):
        return f"{self.__class__.__name__}(queue_wrapper={self.queue_wrapper!r})"

    def add_log_cmd(self, print_str: str):
        self.queue_wrapper.queue.append(LogQueueEntry(print_str))

    def add_pus_tc(self, pus_tc: PusTelecommand):
        self.queue_wrapper.queue.append(PusTcEntry(pus_tc))

    def add_ccsds_tc(self, space_packet: SpacePacket):
        self.queue_wrapper.queue.append(SpacePacketEntry(space_packet))

    def add_raw_tc(self, tc: bytes):
        self.queue_wrapper.queue.append(RawTcEntry(tc))

    def add_wait(self, wait_time: timedelta):
        self.queue_wrapper.queue.append(WaitEntry(wait_time))

    def add_wait_ms(self, wait_ms: int):
        self.queue_wrapper.queue.append(WaitEntry.from_millis(wait_ms))

    def add_packet_delay(self, delay: timedelta):
        self.queue_wrapper.queue.append(PacketDelayEntry(delay))

    def add_packet_delay_ms(self, delay_ms: int):
        self.queue_wrapper.queue.append(PacketDelayEntry.from_millis(delay_ms))



import enum
import abc
import struct
from typing import Optional, List

from .filestore import VirtualFilestore
from .mib import LocalEntityCfg
from tmtccmd.logging import get_console_logger
from tmtccmd.com_if import ComInterface
from spacepackets.cfdp.pdu.metadata import MetadataPdu
from spacepackets.cfdp.conf import PduConfig
from spacepackets.cfdp.definitions import (
    TransmissionModes,
    ChecksumTypes,
    SegmentationControl,
    Direction,
    CrcFlag,
    FileSize,
)
from spacepackets.cfdp.tlv import (
    FaultHandlerOverrideTlv,
    FlowLabelTlv,
    MessageToUserTlv,
    FileStoreRequestTlv,
)

LOGGER = get_console_logger()


class CfdpRequest(enum.Enum):
    PUT = 0
    REPORT = 1
    CANCEL = 2
    SUSPEND = 3
    RESUME = 4


class CfdpIndication(enum.Enum):
    TRANSACTION = 0
    EOF = 1
    FINISHED = 2
    METADATA = 3
    FILE_SEGMENT_RECV = 4
    REPORT = 5
    SUSPENDED = 6
    RESUMED = 7
    FAULT = 8
    ABANDONED = 9
    EOF_RECV = 10


class PutRequest:
    destination_id: bytes
    source_file: str
    dest_file: str
    seg_ctrl: SegmentationControl
    fault_handler_overrides: Optional[FaultHandlerOverrideTlv] = None
    flow_label_tlv: Optional[FlowLabelTlv] = None
    trans_mode: TransmissionModes
    closure_requested: bool
    msgs_to_user: Optional[List[MessageToUserTlv]] = None
    fs_requests: Optional[List[FileStoreRequestTlv]] = None


class CfdpStates(enum.Enum):
    IDLE = 0
    CRC_PROCEDURE = 1
    SENDING_METADATA = 2
    SENDING_FILE_DATA_PDUS = 3
    SENDING_EOF_DATA_PDU = 4
    SENDING_FINISH_PDU = 5
    SEND_ACK_PDU = 6


class ByteFlowControl:
    period: float
    max_bytes: int


class BusyError(Exception):
    pass


class SequenceNumberOverflow(Exception):
    pass


class CfdpUserBase:
    def __init__(self, vfs: VirtualFilestore):
        self.vfs = vfs

    @abc.abstractmethod
    def transaction_indication(self, code: CfdpIndication):
        LOGGER.info(f"Received transaction indication {code}")


class CfdpHandler:
    def __init__(
        self,
        cfg: LocalEntityCfg,
        com_if: Optional[ComInterface],
        cfdp_user: CfdpUserBase,
        byte_flow_ctrl: ByteFlowControl,
    ):
        """

        :param cfg: Local entity configuration
        :param com_if: Communication interface used to send messages
        :param cfdp_user: CFDP user which will receive indication messages and which also contains
            the virtual filestore implementation
        :param byte_flow_ctrl: Controls the number of bytes sent in a certain interval
            The state machine will only send packets if the maximum number of specified bytes
            is not exceeded in the specified time interval
        """
        # The ID is going to be constant after initialization, store in separately
        self.id = cfg.local_entity_id
        self.cfg = cfg
        self.com_if = com_if
        self.cfdp_user = cfdp_user
        self.state = CfdpStates.IDLE
        self.seq_num = 0
        self.byte_flow_ctrl = byte_flow_ctrl

        self.__current_put_request: Optional[PutRequest] = None

    @property
    def com_if(self):
        return self.__com_if

    @com_if.setter
    def com_if(self, com_if: ComInterface):
        self.__com_if = com_if

    def state_machine(self):
        """Perform the CFDP state machine

        :raises SequenceNumberOverflow: Overflow of sequence number occured. In this case, the
            number will be reset but no operation will occured and the state machine needs
            to be called again
        """
        if self.state != CfdpStates.IDLE:
            if self.state == CfdpStates.CRC_PROCEDURE:
                # Skip this step for now
                self.state = CfdpStates.SENDING_METADATA
            if self.state == CfdpStates.SENDING_METADATA:
                # TODO: CRC flag is derived from remote entity ID configuration
                # TODO: Determine file size and check whether source file is valid
                pdu_conf = PduConfig(
                    seg_ctrl=self.__current_put_request.seg_ctrl,
                    dest_entity_id=self.__current_put_request.destination_id,
                    source_entity_id=self.id,
                    crc_flag=CrcFlag.GLOBAL_CONFIG,
                    direction=Direction.TOWARDS_RECEIVER,
                    transaction_seq_num=self.__get_next_seq_num(),
                    file_size=FileSize.GLOBAL_CONFIG,
                    trans_mode=self.__current_put_request.trans_mode,
                )
                self.send_metadata_pdu(
                    pdu_conf=pdu_conf,
                    dest_file=self.__current_put_request.dest_file,
                    source_file=self.__current_put_request.source_file,
                    closure_requested=False,
                )
                self.state = CfdpStates.SENDING_FILE_DATA_PDUS
            pass

    def __get_next_seq_num(self) -> bytes:
        if self.cfg.length_seq_num == 1:
            if self.seq_num == pow(2, 8) - 1:
                LOGGER.warning("8-bit transaction sequence number overflowed!")
                self.seq_num = 0
                raise SequenceNumberOverflow
            self.seq_num += 1
            return bytes([self.seq_num])
        elif self.cfg.length_seq_num == 2:
            if self.seq_num == pow(2, 16) - 1:
                LOGGER.warning("16-bit transaction sequence number overflowed!")
                self.seq_num = 0
                raise SequenceNumberOverflow
            return struct.pack("!H", self.seq_num)
        elif self.cfg.length_seq_num == 4:
            if self.seq_num == pow(2, 32) - 1:
                LOGGER.warning("32-bit transaction sequence number overflowed!")
                self.seq_num = 0
                raise SequenceNumberOverflow
            return struct.pack("!I", self.seq_num)

    def pass_packet(self, raw_tm_packet: bytes):
        pass

    def put_request(self, put_request: PutRequest):
        """A put request initiates a copy procedure. For now, only one put request at a time
        is allowed"""
        if self.state != CfdpStates.IDLE:
            raise BusyError
        self.__current_put_request = put_request
        self.state = CfdpStates.CRC_PROCEDURE

    def send_metadata_pdu(
        self,
        pdu_conf: PduConfig,
        source_file: str,
        dest_file: str,
        closure_requested: bool,
    ):
        metadata_pdu = MetadataPdu(
            pdu_conf=pdu_conf,
            file_size=0,
            source_file_name=source_file,
            dest_file_name=dest_file,
            checksum_type=ChecksumTypes.NULL_CHECKSUM,
            closure_requested=closure_requested,
        )
        data = metadata_pdu.pack()
        self.com_if.send(data=data)



"""
@brief      Core components for mode commanding (custom PUS service)
"""
import enum
import struct
from tmtccmd.pus.pus_200_fsfw_mode import Subservices


class Modes(enum.IntEnum):
    OFF = 0
    ON = 1
    NORMAL = 2
    RAW = 3


def pack_mode_data(object_id: bytes, mode: Modes, submode: int) -> bytearray:
    """Mode 0: Off, Mode 1: Mode On, Mode 2: Mode Normal, Mode 3: Mode Raw"""
    mode_data = bytearray()
    mode_data += object_id + struct.pack("!I", mode) + struct.pack("B", submode)
    return mode_data



import os
from pathlib import Path
import platform
from unittest import TestCase

from tmtccmd.pus.seqcnt import FileSeqCountProvider
from tempfile import NamedTemporaryFile


class TestSeqCount(TestCase):
    def setUp(self) -> None:
        self.file_name = Path("seq_cnt.txt")

    def test_basic(self):
        if platform.system() != "Windows":
            with NamedTemporaryFile("w+t") as file:
                file.write("0\n")
                file.seek(0)
                seq_cnt_provider = FileSeqCountProvider(Path(file.name))
                seq_cnt = seq_cnt_provider.current()
                self.assertEqual(seq_cnt, 0)
                self.assertEqual(next(seq_cnt_provider), 1)
                self.assertEqual(seq_cnt_provider.next_seq_count(), 2)
                file.seek(0)
                file.write(f"{pow(2, 14) - 1}\n")
                file.flush()
                # Assert rollover
                self.assertEqual(next(seq_cnt_provider), 0)

    def test_with_real_file(self):
        seq_cnt_provider = FileSeqCountProvider(self.file_name)
        self.assertTrue(self.file_name.exists())
        self.assertEqual(seq_cnt_provider.current(), 0)
        self.assertEqual(next(seq_cnt_provider), 1)
        pass

    def test_file_deleted_runtime(self):
        seq_cnt_provider = FileSeqCountProvider(self.file_name)
        self.assertTrue(self.file_name.exists())
        os.remove(self.file_name)
        with self.assertRaises(FileNotFoundError):
            next(seq_cnt_provider)
        with self.assertRaises(FileNotFoundError):
            seq_cnt_provider.current()

    def test_faulty_file_entry(self):
        if platform.system() != "Windows":
            with NamedTemporaryFile("w+t") as file:
                file.write("-1\n")
                file.seek(0)
                seq_cnt_provider = FileSeqCountProvider(Path(file.name))
                with self.assertRaises(ValueError):
                    next(seq_cnt_provider)
                file.write(f"{pow(2, 15)}\n")
                file.seek(0)
                file.flush()
                seq_cnt_provider = FileSeqCountProvider(Path(file.name))
                with self.assertRaises(ValueError):
                    next(seq_cnt_provider)

    def tearDown(self) -> None:
        if self.file_name.exists():
            os.remove(self.file_name)



from __future__ import annotations

import time
from typing import Optional
from datetime import timedelta


def time_ms() -> int:
    return round(time.time() * 1000)


class Countdown:
    def __init__(self, init_timeout_secs: Optional[timedelta]):
        if init_timeout_secs is not None:
            self._timeout_ms = int(init_timeout_secs / timedelta(milliseconds=1))
            self._start_time_ms = time_ms()
        else:
            self._timeout_ms = 0
            self._start_time_ms = 0

    @classmethod
    def from_millis(cls, timeout_ms: int) -> Countdown:
        return cls(timedelta(milliseconds=timeout_ms))

    @property
    def timeout(self):
        return self._timeout_ms

    @timeout.setter
    def timeout(self, timeout: timedelta):
        self._timeout_ms = round(timeout / timedelta(milliseconds=1))

    def timed_out(self) -> bool:
        if round(time_ms() - self._start_time_ms) >= self._timeout_ms:
            return True
        else:
            return False

    def busy(self) -> bool:
        return not self.timed_out()

    def reset(self, new_timeout: Optional[timedelta] = None):
        if new_timeout is not None:
            self.timeout = new_timeout
        self.start()

    def start(self):
        self._start_time_ms = time_ms()

    def time_out(self):
        self._start_time_ms = 0

    def rem_time(self) -> timedelta:
        end_time = self._start_time_ms + self._timeout_ms
        current = time_ms()
        if end_time < current:
            return timedelta()
        return timedelta(milliseconds=end_time - current)



from abc import abstractmethod
import enum
from datetime import timedelta
from typing import Optional

from tmtccmd.tc.ccsds_seq_sender import SeqResultWrapper, SenderMode


class TcMode(enum.IntEnum):
    IDLE = 0
    ONE_QUEUE = 1
    MULTI_QUEUE = 2


class TmMode(enum.IntEnum):
    IDLE = 0
    LISTENER = 1


class ModeWrapper:
    def __init__(self):
        self.tc_mode = TcMode.IDLE
        self.tm_mode = TmMode.IDLE

    def __str__(self):
        return f"{self.__class__.__name__}: tc_mode={self.tc_mode!r}, tm_mode={self.tm_mode!r}"


class BackendRequest(enum.IntEnum):
    """These requests can be treated like recommendations on what to do after calling the backend
    handler functions and the :py:meth:`BackendState.mode_to_req` function.

    Brief explanation of fields:
     1. NONE: No special recommendation
     2. TERMINATION_NO_ERROR: Will be returned for the One Queue mode after finishing queue handling.
     3. DELAY_IDLE: TC and TM mode are idle, so there is nothing to do
     4. DELAY_LISTENER: TC handling is not active but TM listening is active. Delay to
        wait for new TM packets
     5. CALL_NEXT: It is recommended to call the handler functions immediately, for example to
        handle the next entry in the TC queue
    """

    NONE = 0
    TERMINATION_NO_ERROR = 1
    DELAY_IDLE = 2
    DELAY_LISTENER = 3
    DELAY_CUSTOM = 4
    CALL_NEXT = 5


class BackendState:
    def __init__(
        self,
        mode_wrapper: ModeWrapper = ModeWrapper(),
        req: BackendRequest = BackendRequest.NONE,
    ):
        self._mode_wrapper = mode_wrapper
        self._req = req
        self._recommended_delay = timedelta()
        self._sender_res = SeqResultWrapper(SenderMode.DONE)

    @property
    def next_delay(self):
        return self._recommended_delay

    @property
    def request(self):
        return self._req

    @property
    def sender_res(self):
        return self._sender_res

    @property
    def tc_mode(self):
        return self._mode_wrapper.tc_mode

    @property
    def tm_mode(self):
        return self._mode_wrapper.tm_mode

    @property
    def mode_wrapper(self):
        return self._mode_wrapper


class BackendController:
    def __init__(self):
        self.next_tc_mode = TcMode.IDLE
        self.next_tm_mode = TmMode.IDLE


class BackendBase:
    @abstractmethod
    def open_com_if(self):
        """Start the backend. Raise RuntimeError on failure"""
        pass

    @abstractmethod
    def close_com_if(self):
        pass

    @abstractmethod
    def periodic_op(self, args: Optional[any]) -> BackendState:
        pass



from dataclasses import dataclass
from typing import Callable
from spacepackets.cfdp.definitions import FaultHandlerCodes

# User can specify a function which takes the fault handler code as an argument and returns nothing
FaultHandlerT = Callable[[FaultHandlerCodes], None]


@dataclass
class LocalEntityCfg:
    local_entity_id: bytes
    eof_sent_indication_required: bool
    eof_recv_indication_required: bool
    file_segment_recvd_required: bool
    transaction_finished_indication_required: bool
    suspended_indication_required: bool
    resumed_indication_required: bool
    default_fault_handlers: FaultHandlerT
    length_seq_num: int = 2
    # I'm just going to assume that 255 possible IDs are sufficient for most applications
    length_entity_ids: int = 1



import os
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from spacepackets.ecss import PusTelecommand
from tmtccmd.com_if.dummy import DummyComIF
from tmtccmd.com_if.utils import determine_com_if


class TestComIF(TestCase):
    def setUp(self) -> None:
        self.json_file = "test.json"

    def test_com_if_utils(self):
        with patch("tmtccmd.com_if.utils.wrapped_prompt", side_effect=["0", "yes"]):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
            with open(self.json_file) as file:
                lines = file.readlines()
                lines[0] = "{\n"
                lines[1] = '    "com_if": "test-com-if"\n'
                lines[2] = "}"
            os.remove(self.json_file)
        with patch("tmtccmd.com_if.utils.wrapped_prompt", side_effect=["0", "no"]):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
            with open(self.json_file) as file:
                lines = file.readlines()
                lines[0] = "{}"
        with patch(
            "tmtccmd.com_if.utils.wrapped_prompt",
            side_effect=["1", "0", "no"],
        ):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")
        with patch(
            "tmtccmd.com_if.utils.wrapped_prompt",
            side_effect=["blub", "0", "no"],
        ):
            test_dict = {"test-com-if": ("Some more info", None)}
            com_if = determine_com_if(test_dict, self.json_file, True)
            self.assertEqual(com_if, "test-com-if")

    def test_dummy_if(self):
        dummy_com_if = DummyComIF()
        self.assertFalse(dummy_com_if.is_open())
        dummy_com_if.open()
        self.assertTrue(dummy_com_if.is_open())
        self.assertFalse(dummy_com_if.initialized)
        dummy_com_if.initialize()
        self.assertTrue(dummy_com_if.initialized)
        self.assertFalse(dummy_com_if.data_available())
        dummy_com_if.send(PusTelecommand(service=17, subservice=1).pack())
        self.assertTrue(dummy_com_if.data_available())
        replies = dummy_com_if.receive()
        # Full verification set (acceptance, start and completion) and ping reply
        self.assertEqual(len(replies), 4)

    def tearDown(self) -> None:
        path = Path(self.json_file)
        if path.exists():
            os.remove(path)



import csv
import os
import copy
from typing import Optional
from tmtccmd.utility.obj_id import ObjectId, ObjectIdDictT
from tmtccmd.pus.pus_5_event import EventInfo, EventDictT
from tmtccmd.utility.retval import RetvalDictT, RetvalInfo


def parse_fsfw_objects_csv(csv_file: str) -> Optional[ObjectIdDictT]:
    if os.path.exists(csv_file):
        obj_id_dict = dict()
        obj_id = ObjectId(object_id=0)
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            for row in csv_reader:
                # Parse hex string
                obj_id.id = int(row[0], 16)
                obj_id.name = row[1]
                obj_id_dict.update({obj_id.as_bytes: copy.copy(obj_id)})
        return obj_id_dict
    else:
        return None


def parse_fsfw_events_csv(csv_file: str) -> Optional[EventDictT]:
    if os.path.exists(csv_file):
        event_dict = dict()
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            info = EventInfo()
            for row in csv_reader:
                info.id = int(row[0])
                info.name = row[2]
                info.severity = row[3]
                info.info = row[4]
                info.file_location = row[5]
                event_dict.update({info.id: copy.copy(info)})
        return event_dict
    else:
        return None


def parse_fsfw_returnvalues_csv(csv_file: str) -> Optional[RetvalDictT]:
    if os.path.exists(csv_file):
        retval_dict = dict()
        with open(csv_file) as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=";")
            info = RetvalInfo()
            for row in csv_reader:
                # Parse hex string
                info.id = int(row[0], 16)
                info.name = row[1]
                info.info = row[2]
                info.if_name = row[5]
                retval_dict.update({info.id: copy.copy(info)})
        return retval_dict
    else:
        return None



from __future__ import annotations

import struct
from abc import abstractmethod
from typing import Deque, Optional

from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.tm import PusVersion, PusTelemetry
from spacepackets.ecss.pus_1_verification import (
    Service1Tm,
    Subservices,
    VerificationParams,
    UnpackParams,
)

from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.logging import get_console_logger

LOGGER = get_console_logger()


class Service1TmExtended(PusTmBase, PusTmInfoBase, Service1Tm):
    """Service 1 TM class representation. Can be used to deserialize raw service 1 packets.
    Only PUS C is supported.
    TODO: Do not use subclassing here, use a wrapper class..
    """

    def __init__(
        self,
        subservice: Subservices,
        verif_params: Optional[VerificationParams] = None,
        time: CdsShortTimestamp = None,
        seq_count: int = 0,
        apid: int = -1,
        packet_version: int = 0b000,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        Service1Tm.__init__(
            self,
            verif_params=verif_params,
            subservice=subservice,
            time=time,
            seq_count=seq_count,
            apid=apid,
            packet_version=packet_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=self.pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=self.pus_tm)
        self._error_param_1 = 0
        self._error_param_2 = 0

    @classmethod
    def __empty(cls) -> Service1TmExtended:
        return cls(subservice=Subservices.INVALID)

    @classmethod
    def unpack(cls, data: bytes, params: UnpackParams) -> Service1TmExtended:
        """Parse a service 1 telemetry packet

        :param params:
        :param data:
        :raises ValueError: Raw telemetry too short
        :return:
        """
        service_1_tm = cls.__empty()
        service_1_tm.pus_tm = PusTelemetry.unpack(raw_telemetry=data)
        cls._unpack_raw_tm(service_1_tm, params)
        # FSFW specific
        if service_1_tm.has_failure_notice:
            service_1_tm._error_param_1 = struct.unpack(
                "!I", service_1_tm.failure_notice.data[0:4]
            )
            service_1_tm._error_param_1 = struct.unpack(
                "!I", service_1_tm.failure_notice.data[4:8]
            )
        return service_1_tm

    @abstractmethod
    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(self.tc_req_id.tc_packet_id)
        content_list.append(self.tc_req_id.tc_psc)
        if self.has_failure_notice:
            if self.is_step_reply:
                content_list.append(str(self.step_id))
            content_list.append(str(hex(self.error_code.val)))
            content_list.append(
                f"hex {self.failure_notice:04x} dec {self.failure_notice}"
            )
            content_list.append(
                f"hex {self.failure_notice.data[0:4]:04x} dec {self._error_param_2}"
            )
        elif self.is_step_reply:
            content_list.append(str(self.step_id))

    @abstractmethod
    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("TC Packet ID")
        header_list.append("TC PSC")
        if self.has_failure_notice:
            if self.is_step_reply:
                header_list.append("Step Number")
            header_list.append("Return Value")
            header_list.append("Error Param 1")
            header_list.append("Error Param 2")
        elif self.is_step_reply:
            header_list.append("Step Number")

    def _unpack_failure_verification(self, params: UnpackParams):
        """Handle parsing a verification failure packet, subservice ID 2, 4, 6 or 8"""
        super()._unpack_failure_verification(params)
        self.set_packet_info("Failure Verficiation")
        subservice = self.pus_tm.subservice
        if subservice == Subservices.TM_ACCEPTANCE_FAILURE:
            self.append_packet_info(" : Acceptance failure")
        elif subservice == Subservices.TM_START_FAILURE:
            self.append_packet_info(" : Start failure")
        elif subservice == Subservices.TM_STEP_FAILURE:
            self.append_packet_info(" : Step Failure")
        elif subservice == Subservices.TM_COMPLETION_FAILURE:
            self.append_packet_info(" : Completion Failure")

    def _unpack_success_verification(self, params: UnpackParams):
        super()._unpack_success_verification(params)
        self.set_packet_info("Success Verification")
        if self.subservice == Subservices.TM_ACCEPTANCE_SUCCESS:
            self.append_packet_info(" : Acceptance success")
        elif self.subservice == Subservices.TM_START_SUCCESS:
            self.append_packet_info(" : Start success")
        elif self.subservice == Subservices.TM_STEP_SUCCESS:
            self.append_packet_info(" : Step Success")
        elif self.subservice == Subservices.TM_COMPLETION_SUCCESS:
            self.append_packet_info(" : Completion success")


PusVerifQueue = Deque[Service1Tm]



from __future__ import annotations
import enum

from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.conf import get_default_tc_apid
from spacepackets.ecss.pus_17_test import Subservices
from tmtccmd.tc.queue import QueueHelper


class CustomSubservices(enum.IntEnum):
    TC_GEN_EVENT = 128


def pack_service_17_ping_command(ssc: int, apid: int = -1) -> PusTelecommand:
    """Generate a simple ping PUS telecommand packet"""
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=17, subservice=Subservices.TC_PING.value, seq_count=ssc, apid=apid
    )


def pack_generic_service17_test(init_ssc: int, q: QueueHelper, apid: int = -1) -> int:
    if apid == -1:
        apid = get_default_tc_apid()
    new_ssc = init_ssc
    q.add_log_cmd("Testing Service 17")
    # ping test
    q.add_log_cmd("Testing Service 17: Ping Test")
    q.add_pus_tc(pack_service_17_ping_command(ssc=new_ssc))
    new_ssc += 1
    # enable event
    q.add_log_cmd("Testing Service 17: Enable Event")
    q.add_pus_tc(PusTelecommand(service=5, subservice=5, seq_count=new_ssc, apid=apid))
    new_ssc += 1
    # test event
    q.add_log_cmd("Testing Service 17: Trigger event")
    q.add_pus_tc(
        PusTelecommand(
            service=17,
            subservice=CustomSubservices.TC_GEN_EVENT,
            seq_count=new_ssc,
            apid=apid,
        )
    )
    new_ssc += 1
    # invalid subservice
    q.add_log_cmd("Testing Service 17: Invalid subservice")
    q.add_pus_tc(
        PusTelecommand(service=17, subservice=243, seq_count=new_ssc, apid=apid)
    )
    new_ssc += 1
    return new_ssc



import time
from datetime import timedelta
from unittest import TestCase
from tmtccmd.utility.countdown import Countdown


class CountdownTest(TestCase):
    def test_basic(self):
        test_cd = Countdown.from_millis(50)
        self.assertTrue(test_cd.busy())
        self.assertFalse(test_cd.timed_out())
        self.assertTrue(test_cd.rem_time().total_seconds() * 1000 > 40)
        time.sleep(0.05)
        self.assertTrue(test_cd.timed_out())
        self.assertTrue(test_cd.rem_time() == timedelta())
        test_cd.timeout = timedelta(seconds=0.1)
        self.assertTrue(test_cd.busy())
        self.assertFalse(test_cd.timed_out())
        time.sleep(0.1)
        self.assertTrue(test_cd.timed_out())
        test_cd.reset(timedelta(seconds=0.5))
        self.assertTrue(test_cd.rem_time().total_seconds() * 1000 > 45)
        self.assertTrue(test_cd.busy())
        self.assertFalse(test_cd.timed_out())
        test_cd.reset(timedelta(milliseconds=50))
        self.assertTrue(test_cd.busy())
        test_cd.time_out()
        self.assertTrue(test_cd.timed_out())



import collections.abc
import enum
import pprint
from typing import Union, List, Dict

from spacepackets.ecss.conf import (
    set_default_tc_apid,
    set_default_tm_apid,
)

from tmtccmd.logging import get_console_logger
from tmtccmd.core.globals_manager import update_global, get_global
from tmtccmd.config import (
    CoreModeList,
    CoreServiceList,
    CoreModeStrings,
    CORE_COM_IF_DICT,
    CoreComInterfaces,
    ComIfDictT,
)
from tmtccmd.config.tmtc import TmTcDefWrapper, OpCodeEntry


LOGGER = get_console_logger()
DEF_WRAPPER = None


class CoreGlobalIds(enum.IntEnum):
    """
    Numbers from 128 to 200 are reserved for core globals
    """

    # Object handles
    TMTC_HOOK = 128
    COM_INTERFACE_HANDLE = 129
    TM_LISTENER_HANDLE = 130
    TMTC_PRINTER_HANDLE = 131
    TM_HANDLER_HANDLE = 132
    PRETTY_PRINTER = 133

    # Parameters
    JSON_CFG_PATH = 139
    MODE = 141
    CURRENT_SERVICE = 142
    COM_IF = 144
    OP_CODE = 145
    TM_TIMEOUT = 146
    SERVICE_OP_CODE_DICT = 147
    COM_IF_DICT = 148

    # Miscellaneous
    DISPLAY_MODE = 150
    USE_LISTENER_AFTER_OP = 151
    PRINT_HK = 152
    PRINT_TM = 153
    PRINT_RAW_TM = 154
    PRINT_TO_FILE = 155
    RESEND_TC = 156
    TC_SEND_TIMEOUT_FACTOR = 157

    # Config dictionaries
    USE_SERIAL = 160
    SERIAL_CONFIG = 161
    USE_ETHERNET = 162
    ETHERNET_CONFIG = 163
    END = 300


def set_json_cfg_path(json_cfg_path: str):
    update_global(CoreGlobalIds.JSON_CFG_PATH, json_cfg_path)


def get_json_cfg_path() -> str:
    return get_global(CoreGlobalIds.JSON_CFG_PATH)


def set_glob_com_if_dict(custom_com_if_dict: ComIfDictT):
    CORE_COM_IF_DICT.update(custom_com_if_dict)
    update_global(CoreGlobalIds.COM_IF_DICT, CORE_COM_IF_DICT)


def get_glob_com_if_dict() -> ComIfDictT:
    return get_global(CoreGlobalIds.COM_IF_DICT)


def set_default_globals_pre_args_parsing(
    apid: int,
    com_if_id: str = CoreComInterfaces.DUMMY.value,
    custom_com_if_dict=None,
    display_mode="long",
    tm_timeout: float = 4.0,
    print_to_file: bool = True,
    tc_send_timeout_factor: float = 2.0,
):
    if custom_com_if_dict is None:
        custom_com_if_dict = dict()
    set_default_tc_apid(tc_apid=apid)
    set_default_tm_apid(tm_apid=apid)
    update_global(CoreGlobalIds.COM_IF, com_if_id)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, tc_send_timeout_factor)
    update_global(CoreGlobalIds.TM_TIMEOUT, tm_timeout)
    update_global(CoreGlobalIds.DISPLAY_MODE, display_mode)
    update_global(CoreGlobalIds.PRINT_TO_FILE, print_to_file)
    update_global(CoreGlobalIds.CURRENT_SERVICE, CoreServiceList.SERVICE_17.value)
    update_global(CoreGlobalIds.SERIAL_CONFIG, dict())
    update_global(CoreGlobalIds.ETHERNET_CONFIG, dict())
    set_glob_com_if_dict(custom_com_if_dict=custom_com_if_dict)
    pp = pprint.PrettyPrinter()
    update_global(CoreGlobalIds.PRETTY_PRINTER, pp)
    update_global(CoreGlobalIds.TM_LISTENER_HANDLE, None)
    update_global(CoreGlobalIds.COM_INTERFACE_HANDLE, None)
    update_global(CoreGlobalIds.TMTC_PRINTER_HANDLE, None)
    update_global(CoreGlobalIds.PRINT_RAW_TM, False)
    update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, True)
    update_global(CoreGlobalIds.RESEND_TC, False)
    update_global(CoreGlobalIds.OP_CODE, "0")
    update_global(CoreGlobalIds.MODE, CoreModeList.LISTENER_MODE)


def handle_mode_arg(
    args,
    custom_modes_list: Union[None, List[Union[collections.abc.Iterable, dict]]] = None,
) -> int:
    # Determine communication interface from arguments. Must be contained in core modes list
    try:
        mode_param = args.mode
    except AttributeError:
        LOGGER.warning("Passed namespace does not contain the mode (-m) argument")
        mode_param = CoreModeList.LISTENER_MODE
    mode_param = check_and_set_core_mode_arg(
        mode_arg=mode_param, custom_modes_list=custom_modes_list
    )
    update_global(CoreGlobalIds.MODE, mode_param)
    return mode_param


def handle_com_if_arg(
    args, json_cfg_path: str, custom_com_if_dict: Dict[str, any] = None
):
    from tmtccmd.com_if.utils import determine_com_if

    all_com_ifs = CORE_COM_IF_DICT
    if custom_com_if_dict is not None:
        all_com_ifs = CORE_COM_IF_DICT.update(custom_com_if_dict)
    try:
        com_if_key = str(args.com_if_id)
    except AttributeError:
        LOGGER.warning("No communication interface specified")
        LOGGER.warning("Trying to set from existing configuration..")
        com_if_key = determine_com_if(
            com_if_dict=all_com_ifs, json_cfg_path=json_cfg_path
        )
    if com_if_key == CoreComInterfaces.UNSPECIFIED.value:
        com_if_key = determine_com_if(
            com_if_dict=all_com_ifs, json_cfg_path=json_cfg_path
        )
    update_global(CoreGlobalIds.COM_IF, com_if_key)
    try:
        LOGGER.info(f"Communication interface: {all_com_ifs[com_if_key]}")
    except KeyError as e:
        LOGGER.error(f"Invalid communication interface key {com_if_key}, error {e}")


def check_and_set_other_args(args):
    if args.listener is not None:
        update_global(CoreGlobalIds.USE_LISTENER_AFTER_OP, args.listener)
    if args.tm_timeout is not None:
        update_global(CoreGlobalIds.TM_TIMEOUT, args.tm_timeout)
    if args.print_hk is not None:
        update_global(CoreGlobalIds.PRINT_HK, args.print_hk)
    if args.print_tm is not None:
        update_global(CoreGlobalIds.PRINT_TM, args.print_tm)
    if args.raw_print is not None:
        update_global(CoreGlobalIds.PRINT_RAW_TM, args.raw_print)
    if args.print_log is not None:
        update_global(CoreGlobalIds.PRINT_TO_FILE, args.print_log)
    if args.resend_tc is not None:
        update_global(CoreGlobalIds.RESEND_TC, args.resend_tc)
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, 3)


def check_and_set_core_mode_arg(
    mode_arg: any,
    custom_modes_list: Union[None, List[Union[dict, collections.abc.Iterable]]] = None,
) -> int:
    from tmtccmd.utility.conf_util import check_args_in_dict

    """Checks whether the mode argument is contained inside the core mode list integer enumeration
    or a custom mode list integer which can be passed optionally.
    This function will set the single command mode as the global mode parameter if the passed mode
    is not found in either enumerations.

    :param mode_arg:
    :param custom_modes_list:
    :return: Mode value which was set
    """
    in_enum, mode_value = check_args_in_dict(
        param=mode_arg, iterable=CoreModeList, warning_hint="mode integers"
    )
    if not in_enum:
        in_enum, mode_value = check_args_in_dict(
            param=mode_arg, iterable=CoreModeStrings, warning_hint="mode strings"
        )
    if in_enum:
        update_global(CoreGlobalIds.MODE, mode_value)
        return mode_value

    mode_arg_invalid = False
    if custom_modes_list is not None:
        for custom_mode_entry in custom_modes_list:
            in_enum, mode_value = check_args_in_dict(
                param=mode_arg, iterable=custom_mode_entry, warning_hint="custom mode"
            )
            if in_enum:
                break
        if not in_enum:
            mode_arg_invalid = True
    else:
        mode_arg_invalid = True

    if mode_arg_invalid:
        LOGGER.warning(
            f"Passed mode argument might be invalid, "
            f"setting to {CoreModeList.ONE_QUEUE_MODE}"
        )
        mode_value = CoreModeList.ONE_QUEUE_MODE
    update_global(CoreGlobalIds.MODE, mode_value)
    return mode_value


def check_and_set_core_service_arg(
    service_arg: any, custom_service_list: collections.abc.Iterable = None
):
    from tmtccmd.utility.conf_util import check_args_in_dict

    in_enum, service_value = check_args_in_dict(
        param=service_arg, iterable=CoreServiceList, warning_hint="service"
    )
    if in_enum:
        update_global(CoreGlobalIds.CURRENT_SERVICE, service_value)
        return

    service_arg_invalid = False
    if custom_service_list is not None:
        for custom_services_entry in custom_service_list:
            in_enum, service_value = check_args_in_dict(
                param=service_arg,
                iterable=custom_services_entry,
                warning_hint="custom mode",
            )
            if in_enum:
                break
        if not in_enum:
            service_arg_invalid = True
    else:
        service_arg_invalid = True

    if service_arg_invalid:
        LOGGER.warning(
            f"Passed service argument might be invalid, "
            f"setting to {CoreServiceList.SERVICE_17}"
        )
        service_value = CoreServiceList.SERVICE_17
    update_global(CoreGlobalIds.CURRENT_SERVICE, service_value)


def get_default_tmtc_defs() -> TmTcDefWrapper:
    global DEF_WRAPPER
    if DEF_WRAPPER is None:
        DEF_WRAPPER = TmTcDefWrapper()
        srv_5 = OpCodeEntry()
        srv_5.add("0", "Event Test")
        DEF_WRAPPER.add_service(
            service_name=CoreServiceList.SERVICE_5.value,
            info="PUS Service 5 Event",
            op_code_entry=srv_5,
        )
        srv_17 = OpCodeEntry()
        srv_17.add("0", "Ping Test")
        DEF_WRAPPER.add_service(
            service_name=CoreServiceList.SERVICE_17.value,
            info="PUS Service 17 Test",
            op_code_entry=srv_17,
        )
    return DEF_WRAPPER



"""Dummy Virtual Communication Interface. Currently serves to use the TMTC program without needing
external hardware or an extra socket
"""
from typing import Optional

from spacepackets.ecss.pus_1_verification import RequestId, VerificationParams
from spacepackets.ecss.tc import PusTelecommand

from tmtccmd.com_if import ComInterface
from tmtccmd.config import CoreComInterfaces
from tmtccmd.tm import TelemetryListT
from tmtccmd.tm.pus_1_verification import Service1TmExtended
from tmtccmd.tm.pus_1_verification import Subservices as Pus1Subservices
from tmtccmd.tm.pus_17_test import Service17TmExtended
from tmtccmd.tm.pus_17_test import Subservices as Pus17Subservices
from tmtccmd.logging import get_console_logger


LOGGER = get_console_logger()


class DummyHandler:
    def __init__(self):
        self.last_tc: Optional[PusTelecommand] = None
        self.next_telemetry_package = []
        self.current_ssc = 0
        self.reply_pending = False

    def pass_telecommand(self, data: bytearray):
        self.last_tc = PusTelecommand.unpack(data)
        self.reply_pending = True
        self.generate_reply_package()

    def generate_reply_package(self):
        """Generate a reply package. Currently, this only generates a reply for a ping telecommand."""
        if self.last_tc.service == 17:
            if self.last_tc.subservice == 1:
                tm_packer = Service1TmExtended(
                    subservice=Pus1Subservices.TM_ACCEPTANCE_SUCCESS,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_ctrl
                        )
                    ),
                )

                self.current_ssc += 1
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                tm_packer = Service1TmExtended(
                    subservice=Pus1Subservices.TM_START_SUCCESS,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_ctrl
                        )
                    ),
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

                tm_packer = Service17TmExtended(subservice=Pus17Subservices.TM_REPLY)
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

                tm_packer = Service1TmExtended(
                    subservice=Pus1Subservices.TM_COMPLETION_SUCCESS,
                    seq_count=self.current_ssc,
                    verif_params=VerificationParams(
                        req_id=RequestId(
                            self.last_tc.packet_id, self.last_tc.packet_seq_ctrl
                        )
                    ),
                )
                tm_packet_raw = tm_packer.pack()
                self.next_telemetry_package.append(tm_packet_raw)
                self.current_ssc += 1

    def receive_reply_package(self) -> TelemetryListT:
        if self.reply_pending:
            return_list = self.next_telemetry_package.copy()
            self.next_telemetry_package.clear()
            self.reply_pending = False
            return return_list
        else:
            return []


class DummyComIF(ComInterface):
    def __init__(self):
        super().__init__(com_if_id=CoreComInterfaces.DUMMY.value)
        self.dummy_handler = DummyHandler()
        self._open = False
        self.initialized = False

    def initialize(self, args: any = None) -> any:
        self.initialized = True

    def open(self, args: any = None) -> None:
        self._open = True

    def is_open(self) -> bool:
        return self._open

    def close(self, args: any = None) -> None:
        self._open = False

    def data_available(self, timeout: float = 0, parameters: any = 0):
        if self.dummy_handler.reply_pending:
            return True
        return False

    def receive(self, parameters: any = 0) -> TelemetryListT:
        return self.dummy_handler.receive_reply_package()

    def send(self, data: bytearray):
        if data is not None:
            self.dummy_handler.pass_telecommand(data)



"""Definitions for the TMTC commander core
"""
from typing import Optional

from tmtccmd.core import ModeWrapper, TmMode, TcMode

from .args import (
    SetupParams,
    create_default_args_parser,
    add_default_tmtccmd_args,
    parse_default_tmtccmd_input_arguments,
    DefProcedureParams,
    ArgParserWrapper,
)
from .defs import (
    CoreModeList,
    CoreModeStrings,
    CoreComInterfaces,
    CORE_COM_IF_DICT,
    default_json_path,
    CoreServiceList,
    ComIfDictT,
)
from .prompt import prompt_op_code, prompt_service
from .tmtc import TmTcDefWrapper
from .hook import TmTcCfgHookBase


def backend_mode_conversion(mode: CoreModeList, mode_wrapper: ModeWrapper):
    if mode == CoreModeStrings[CoreModeList.LISTENER_MODE]:
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.IDLE
    elif mode == CoreModeStrings[CoreModeList.ONE_QUEUE_MODE]:
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.ONE_QUEUE
    elif mode == CoreModeStrings[CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE]:
        mode_wrapper.tc_mode = TcMode.MULTI_QUEUE
        mode_wrapper.tm_mode = TmMode.LISTENER


def get_global_hook_obj() -> Optional[TmTcCfgHookBase]:
    """This function can be used to get the handle to the global hook object.
    :return:
    """
    from tmtccmd import get_console_logger

    logger = get_console_logger()
    try:
        from tmtccmd.core.globals_manager import get_global
        from tmtccmd.config.definitions import CoreGlobalIds

        from typing import cast

        hook_obj_raw = get_global(CoreGlobalIds.TMTC_HOOK)
        if hook_obj_raw is None:
            logger.error("Hook object is invalid!")
            return None
        return cast(TmTcCfgHookBase, hook_obj_raw)
    except ImportError:
        logger.exception("Issues importing modules to get global hook handle!")
        return None
    except AttributeError:
        logger.exception("Attribute error when trying to get global hook handle!")
        return None


class SetupWrapper:
    """This class encapsulates various important setup parameters required by tmtccmd components"""

    def __init__(
        self,
        hook_obj: TmTcCfgHookBase,
        setup_params: SetupParams,
        json_cfg_path: Optional[str] = None,
    ):
        """
        :param hook_obj: User hook object. Needs to be implemented by the user
        :param setup_params: Optional helper wrapper which contains CLI arguments.
        """
        self.hook_obj = hook_obj
        self.json_cfg_path = json_cfg_path
        self._params = setup_params
        self.json_cfg_path = json_cfg_path
        if json_cfg_path is None:
            self.json_cfg_path = default_json_path()

    @property
    def params(self):
        return self._params



import abc
import os
import shutil
import platform
from pathlib import Path

from tmtccmd.logging import get_console_logger
from spacepackets.cfdp.tlv import FilestoreResponseStatusCode

LOGGER = get_console_logger()

FilestoreResult = FilestoreResponseStatusCode


class VirtualFilestore:
    @abc.abstractmethod
    def append_data_to_file(
        self, file: Path, offset: int, data: bytes
    ) -> FilestoreResponseStatusCode:
        """This is not used as part of a filestore request, it is used to build up the received
        file"""
        LOGGER.warning("Appending to file not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def create_file(self, file: Path) -> FilestoreResponseStatusCode:
        LOGGER.warning("Creating file not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def delete_file(self, file: Path) -> FilestoreResponseStatusCode:
        LOGGER.warning("Deleting file not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def rename_file(
        self, _old_file: Path, _new_file: Path
    ) -> FilestoreResponseStatusCode:
        LOGGER.warning("Renaming file not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def replace_file(
        self, _replaced_file: Path, _source_file: Path
    ) -> FilestoreResponseStatusCode:
        LOGGER.warning("Replacing file not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def create_directory(self, _dir_name: Path) -> FilestoreResponseStatusCode:
        LOGGER.warning("Creating directory not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def remove_directory(
        self, _dir_name: Path, recursive: bool
    ) -> FilestoreResponseStatusCode:
        LOGGER.warning("Removing directory not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED

    @abc.abstractmethod
    def list_directory(
        self, _dir_name: Path, _file_name: Path, _recursive: bool = False
    ) -> FilestoreResponseStatusCode:
        LOGGER.warning("Listing directory not implemented in virtual filestore")
        return FilestoreResponseStatusCode.NOT_PERFORMED


class HostFilestore(VirtualFilestore):
    def __init__(self):
        pass

    @abc.abstractmethod
    def append_data_to_file(
        self, file: Path, offset: int, data: bytes
    ) -> FilestoreResponseStatusCode:
        """Primary function used to perform the CFDP Copy Procedure. This will also create a new
        file as long as no other file with the same name exists

        :return:
         - FilestoreResponseStatusCode.APPEND_FROM_DATA_FILE_NOT_EXISTS: File does not exist yet
         - FilestoreResponseStatusCode.APPEND_FROM_DATA_INVALID_OFFSET: Invalid offset
        """
        if not file.exists():
            return FilestoreResponseStatusCode.APPEND_FROM_DATA_FILE_NOT_EXISTS
        with open(file, "r+b") as of:
            file_size = file.stat().st_size
            if offset > file_size:
                return FilestoreResponseStatusCode.APPEND_FROM_DATA_INVALID_OFFSET
            of.seek(offset)
            of.write(data)
        return FilestoreResponseStatusCode.SUCCESS

    def create_file(self, file: Path) -> FilestoreResponseStatusCode:
        """Returns CREATE_NOT_ALLOWED if the file already exists"""
        if file.exists():
            LOGGER.warning("File already exists")
            return FilestoreResponseStatusCode.CREATE_NOT_ALLOWED
        try:
            file = open(file, "x")
            file.close()
            return FilestoreResponseStatusCode.CREATE_SUCCESS
        except OSError:
            LOGGER.exception(f"Creating file {file} failed")
            return FilestoreResponseStatusCode.CREATE_NOT_ALLOWED

    def delete_file(self, file: Path) -> FilestoreResponseStatusCode:
        if not file.exists():
            return FilestoreResponseStatusCode.DELETE_FILE_DOES_NOT_EXIST
        if file.is_dir():
            return FilestoreResponseStatusCode.DELETE_NOT_ALLOWED
        os.remove(file)
        return FilestoreResponseStatusCode.DELETE_SUCCESS

    def rename_file(
        self, old_file: Path, new_file: Path
    ) -> FilestoreResponseStatusCode:
        if old_file.is_dir() or new_file.is_dir():
            LOGGER.exception(f"{old_file} or {new_file} is a directory")
            return FilestoreResponseStatusCode.RENAME_NOT_PERFORMED
        if not old_file.exists():
            return FilestoreResponseStatusCode.RENAME_OLD_FILE_DOES_NOT_EXIST
        if new_file.exists():
            return FilestoreResponseStatusCode.RENAME_NEW_FILE_DOES_EXIST
        old_file.rename(new_file)
        return FilestoreResponseStatusCode.RENAME_SUCCESS

    def replace_file(
        self, replaced_file: Path, source_file: Path
    ) -> FilestoreResponseStatusCode:
        if replaced_file.is_dir() or source_file.is_dir():
            LOGGER.warning(f"{replaced_file} is a directory")
            return FilestoreResponseStatusCode.REPLACE_NOT_ALLOWED
        if not replaced_file.exists():
            return (
                FilestoreResponseStatusCode.REPLACE_FILE_NAME_ONE_TO_BE_REPLACED_DOES_NOT_EXIST
            )
        if not source_file.exists():
            return (
                FilestoreResponseStatusCode.REPLACE_FILE_NAME_TWO_REPLACE_SOURCE_NOT_EXIST
            )
        source_file.replace(replaced_file)

    def remove_directory(
        self, dir_name: Path, recursive: bool = False
    ) -> FilestoreResponseStatusCode:
        if not dir_name.exists():
            LOGGER.warning(f"{dir_name} does not exist")
            return FilestoreResponseStatusCode.REMOVE_DIR_DOES_NOT_EXIST
        elif not dir_name.is_dir():
            LOGGER.warning(f"{dir_name} is not a directory")
            return FilestoreResponseStatusCode.REMOVE_DIR_NOT_ALLOWED
        if recursive:
            shutil.rmtree(dir_name)
        else:
            try:
                os.rmdir(dir_name)
                return FilestoreResponseStatusCode.REMOVE_DIR_SUCCESS
            except OSError:
                LOGGER.exception(f"Removing directory {dir_name} failed")
                return FilestoreResponseStatusCode.RENAME_NOT_PERFORMED

    def create_directory(self, dir_name: Path) -> FilestoreResponseStatusCode:
        if dir_name.exists():
            # It does not really matter if the existing structure is a file or a directory
            return FilestoreResponseStatusCode.CREATE_DIR_CAN_NOT_BE_CREATED
        os.mkdir(dir_name)
        return FilestoreResponseStatusCode.CREATE_DIR_SUCCESS

    def list_directory(
        self, dir_name: Path, target_file: Path, recursive: bool = False
    ) -> FilestoreResponseStatusCode:
        """List a directory

        :param dir_name: Name of directory to list
        :param target_file: The list will be written into this target file
        :param recursive:
        :return:
        """
        if target_file.exists():
            open_flag = "a"
        else:
            open_flag = "w"
        with open(target_file, open_flag) as of:
            if platform.system() == "Linux" or platform.system() == "Darwin":
                cmd = "ls -al"
            elif platform.system() == "Windows":
                cmd = "dir"
            else:
                LOGGER.warning(
                    f"Unknown OS {platform.system()}, do not know how to list directory"
                )
                return FilestoreResponseStatusCode.NOT_PERFORMED
            of.write(f"Contents of directory {dir_name} generated with '{cmd}':\n")
            curr_path = os.getcwd()
            os.chdir(dir_name)
            os.system(f"{cmd} >> {target_file}")
            os.chdir(curr_path)
        return FilestoreResponseStatusCode.SUCCESS



"""Contains definitions and functions related to PUS Service 3 Telecommands.
"""
import struct

from spacepackets.ecss.tc import PusTelecommand
from spacepackets.ecss.pus_3_hk import Subservices


def make_sid(object_id: bytes, set_id: int) -> bytearray:
    sid_raw = bytearray()
    sid_raw += object_id + struct.pack(">I", set_id)
    return sid_raw


def make_interval(interval_seconds: float) -> bytearray:
    return bytearray(struct.pack("!f", interval_seconds))


def enable_periodic_hk_command(diag: bool, sid: bytes, ssc: int) -> PusTelecommand:
    return __generate_periodic_hk_command(diag=diag, enable=True, sid=sid, ssc=ssc)


def enable_periodic_hk_command_with_interval(
    diag: bool, sid: bytes, interval_seconds: float, ssc: int
) -> (PusTelecommand, PusTelecommand):
    cmd0 = modify_collection_interval(diag, sid, interval_seconds, ssc)
    cmd1 = __generate_periodic_hk_command(diag=diag, enable=True, sid=sid, ssc=ssc)
    return cmd0, cmd1


def disable_periodic_hk_command(diag: bool, sid: bytes, ssc: int) -> PusTelecommand:
    return __generate_periodic_hk_command(diag=diag, enable=False, sid=sid, ssc=ssc)


def __generate_periodic_hk_command(
    diag: bool, enable: bool, sid: bytes, ssc: int
) -> PusTelecommand:
    app_data = bytearray(sid)
    if enable:
        if diag:
            subservice = Subservices.TC_ENABLE_PERIODIC_DIAGNOSTICS_GEN
        else:
            subservice = Subservices.TC_ENABLE_PERIODIC_HK_GEN
    else:
        if diag:
            subservice = Subservices.TC_DISABLE_PERIODIC_DIAGNOSTICS_GEN
        else:
            subservice = Subservices.TC_DISABLE_PERIODIC_HK_GEN
    return PusTelecommand(
        service=3, subservice=subservice, seq_count=ssc, app_data=app_data
    )


def modify_collection_interval(
    diag: bool, sid: bytes, interval_seconds: float, ssc: int
) -> PusTelecommand:
    app_data = bytearray(sid)
    app_data += make_interval(interval_seconds)
    if diag:
        subservice = Subservices.TC_MODIFY_DIAGNOSTICS_REPORT_COLLECTION_INTERVAL
    else:
        subservice = Subservices.TC_MODIFY_PARAMETER_REPORT_COLLECTION_INTERVAL
    return PusTelecommand(
        service=3, subservice=subservice, seq_count=ssc, app_data=app_data
    )


def generate_one_hk_command(sid: bytes, ssc: int) -> PusTelecommand:
    return PusTelecommand(
        service=3,
        subservice=Subservices.TC_GENERATE_ONE_PARAMETER_REPORT,
        seq_count=ssc,
        app_data=sid,
    )


def generate_one_diag_command(sid: bytes, ssc: int) -> PusTelecommand:
    return PusTelecommand(
        service=3,
        subservice=Subservices.TC_GENERATE_ONE_DIAGNOSTICS_REPORT,
        seq_count=ssc,
        app_data=sid,
    )



"""Contains definitions and functions related to PUS Service 20 Telecommands.
"""
import struct
from typing import Optional

from spacepackets.ecss.fields import Ptc, PfcUnsigned, PfcReal
from tmtccmd.pus.pus_20_params import (
    CustomSubservices,
)
from spacepackets.ecss.tc import PusTelecommand
from tmtccmd.logging import get_console_logger

logger = get_console_logger()


def pack_fsfw_load_param_cmd(
    app_data: bytes, ssc: int, apid: int = -1
) -> PusTelecommand:
    return PusTelecommand(
        service=20,
        subservice=CustomSubservices.LOAD,
        app_data=app_data,
        apid=apid,
        seq_count=ssc,
    )


def pack_boolean_parameter_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: bool
) -> Optional[bytearray]:
    """Generic function to pack a the application data for a parameter service command.
    Tailored towards FSFW applications.
    :param object_id:
    :param domain_id:
    :param unique_id:
    :param parameter:
    :return: Application data
    """
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.UNSIGNED,
        pfc=PfcUnsigned.ONE_BYTE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.append(parameter)
    return data_to_pack


def pack_scalar_double_param_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Optional[bytearray]:
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.REAL,
        pfc=PfcReal.DOUBLE_PRECISION_IEEE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.extend(struct.pack("!d", parameter))
    return data_to_pack


def pack_scalar_float_param_app_data(
    object_id: bytes, domain_id: int, unique_id: int, parameter: float
) -> Optional[bytearray]:
    data_to_pack = prepare_param_packet_header(
        object_id=object_id,
        domain_id=domain_id,
        unique_id=unique_id,
        ptc=Ptc.REAL,
        pfc=PfcReal.FLOAT_SIMPLE_PRECISION_IEEE,
        rows=1,
        columns=1,
    )
    if data_to_pack is not None:
        data_to_pack.extend(struct.pack("!f", parameter))
    return data_to_pack


def prepare_param_packet_header(
    object_id: bytes,
    domain_id: int,
    unique_id: int,
    ptc: Ptc,
    pfc: int,
    rows: int,
    columns: int,
    start_at_idx: int = 0,
) -> Optional[bytearray]:
    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    if unique_id > 255:
        logger.warning("Invalid unique ID, should be smaller than 255!")
        return None
    parameter_id[1] = unique_id
    parameter_id[2] = (start_at_idx >> 8) & 0xFF
    parameter_id[3] = start_at_idx & 0xFF
    data_to_pack = bytearray(object_id)
    data_to_pack.extend(parameter_id)
    data_to_pack.extend(
        pack_type_and_matrix_data(ptc=ptc, pfc=pfc, rows=rows, columns=columns)
    )
    return data_to_pack


def pack_type_and_matrix_data(ptc: int, pfc: int, rows: int, columns: int) -> bytearray:
    # noinspection PyPep8
    """Packs the parameter information field, which contains the ECSS PTC and PFC numbers and the
    number of columns and rows in the parameter.
    See https://ecss.nl/standard/ecss-e-st-70-41c-space-engineering-telemetry-and-telecommand-packet-utilization-15-april-2016/
    p.428 for more information.

    :param ptc:     ECSS PTC number
    :param pfc:     ECSS PFC number
    :param rows:     Number of rows in parameter (for matrix entries, 1 for vector entries,
        1 for scalar entries)
    :param columns:  Number of columns in parameter (for matrix or vector entries,
        1 for scalar entries)
    :return: Parameter information field as 4 byte bytearray
    """
    data = bytearray(4)
    data[0] = ptc
    data[1] = pfc
    data[2] = rows
    data[3] = columns
    return data


def pack_parameter_id(domain_id: int, unique_id: int, linear_index: int) -> bytearray:
    """Packs the Parameter ID (bytearray with 4 bytes) which is part of the service 20 packets.
    The first byte of the parameter ID is the domain ID, the second byte is a unique ID and the
    last two bytes are a linear index if a parameter is not loaded from index 0.
    :param domain_id:       One byte domain ID
    :param unique_id:       One byte unique ID
    :param linear_index:    Two byte linear index.
    """
    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    parameter_id[1] = unique_id
    parameter_id[2] = linear_index >> 8 & 0xFF
    parameter_id[3] = linear_index & 0xFF
    return parameter_id



from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.conf import FETCH_GLOBAL_APID
from tmtccmd.pus import PusServices
from tmtccmd.pus.pus_11_tc_sched import Subservices


def __generic_param_less_tc_sched_cmd(
    subservice: int, ssc: int, apid: int = -1
) -> PusTelecommand:
    return PusTelecommand(
        service=PusServices.S11_TC_SCHED,
        subservice=subservice,
        seq_count=ssc,
        apid=apid,
    )


def generate_enable_tc_sched_cmd(
    ssc: int, apid: int = FETCH_GLOBAL_APID
) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservices.TC_ENABLE, ssc=ssc, apid=apid
    )


def generate_disable_tc_sched_cmd(
    ssc: int, apid: int = FETCH_GLOBAL_APID
) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservices.TC_DISABLE, ssc=ssc, apid=apid
    )


def generate_reset_tc_sched_cmd(
    ssc: int, apid: int = FETCH_GLOBAL_APID
) -> PusTelecommand:
    return __generic_param_less_tc_sched_cmd(
        subservice=Subservices.TC_RESET, ssc=ssc, apid=apid
    )


def generate_time_tagged_cmd(
    release_time: bytes,
    tc_to_insert: PusTelecommand,
    ssc: int,
    apid: int = FETCH_GLOBAL_APID,
):
    return PusTelecommand(
        service=PusServices.S11_TC_SCHED,
        subservice=Subservices.TC_INSERT,
        app_data=pack_time_tagged_tc_app_data(release_time, tc_to_insert),
        seq_count=ssc,
        apid=apid,
    )


def pack_time_tagged_tc_app_data(
    release_time: bytes, tc_to_insert: PusTelecommand
) -> bytes:
    """This function packs another TC into an insert activity TC[11,4]
    :param release_time: Absolute time when TC shall be released/run
    :param tc_to_insert: The TC which shall be inserted
    """
    app_data = bytearray()
    # pack the release time
    app_data.extend(release_time)
    # followed by the tc
    app_data.extend(tc_to_insert.pack())
    return app_data



"""Contains definitions and functions related to PUS Service 5 Telecommands.
"""
from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.conf import get_default_tc_apid
from spacepackets.ecss.pus_5_event import Subservices

from tmtccmd.tc.queue import QueueHelper


def pack_enable_event_reporting_command(ssc: int, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=5,
        subservice=Subservices.TC_ENABLE_EVENT_REPORTING,
        seq_count=ssc,
        apid=apid,
    )


def pack_disable_event_reporting_command(ssc: int, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=5,
        subservice=Subservices.TC_DISABLE_EVENT_REPORTING,
        seq_count=ssc,
        apid=apid,
    )


def pack_generic_service5_test_into(q: QueueHelper, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    q.add_log_cmd("Testing Service 5")
    # invalid subservice
    q.add_log_cmd("Testing Service 5: Invalid subservice")
    q.add_pus_tc(PusTelecommand(service=5, subservice=1, apid=apid, seq_count=500))
    # disable events
    q.add_log_cmd("Testing Service 5: Disable event")
    q.add_pus_tc(pack_disable_event_reporting_command(ssc=501))
    # trigger event
    q.add_log_cmd("Testing Service 5: Trigger event")
    q.add_pus_tc(PusTelecommand(service=17, subservice=128, apid=apid, seq_count=510))
    # enable event
    q.add_log_cmd("Testing Service 5: Enable event")
    q.add_pus_tc(pack_enable_event_reporting_command(ssc=520))
    # trigger event
    q.add_log_cmd("Testing Service 5: Trigger another event")
    q.add_pus_tc(PusTelecommand(service=17, subservice=128, apid=apid, seq_count=530))



from typing import Union, List, Optional, Dict, Tuple

ServiceNameT = str
ServiceInfoT = str
OpCodeNameT = Union[str, List[str]]
OpCodeInfoT = str


class OpCodeOptionBase:
    def __init__(self):
        pass


OpCodeDictT = Dict[str, Tuple[OpCodeInfoT, OpCodeOptionBase]]


class OpCodeEntry:
    def __init__(self, init_dict: Optional[OpCodeDictT] = None):
        if init_dict is not None:
            self._op_code_dict = init_dict
        else:
            self._op_code_dict: OpCodeDictT = dict()

    def add(
        self,
        keys: OpCodeNameT,
        info: str,
        options: OpCodeOptionBase = OpCodeOptionBase(),
    ):
        if isinstance(keys, str):
            keys = [keys]
        self._op_code_dict.update(OpCodeDictT.fromkeys(keys, (info, options)))

    def sort(self):
        self._op_code_dict = {
            key: self._op_code_dict[key] for key in sorted(self._op_code_dict.keys())
        }

    def info(self, op_code: str) -> Optional[str]:
        entry_tuple = self._op_code_dict.get(op_code)
        if entry_tuple is not None:
            return entry_tuple[0]

    def __repr__(self):
        return f"{self.__class__.__name__}(init_dict={self._op_code_dict!r}"

    @property
    def op_code_dict(self):
        return self._op_code_dict


# It is possible to specify a service without any op codes
ServiceDictValueT = Optional[Tuple[ServiceInfoT, OpCodeEntry]]
ServiceOpCodeDictT = Dict[ServiceNameT, ServiceDictValueT]


class TmTcDefWrapper:
    def __init__(self, init_defs: Optional[ServiceOpCodeDictT] = None):
        if init_defs is None:
            self.defs: ServiceOpCodeDictT = dict()
        else:
            self.defs = init_defs

    def __repr__(self):
        return f"{self.__class__.__name__}(init_defs={self.defs!r}"

    def add_service(
        self,
        service_name: str,
        info: str,
        op_code_entry: OpCodeEntry,
    ):
        self.defs.update({service_name: (info, op_code_entry)})

    def op_code_entry(self, service_name: str) -> Optional[OpCodeEntry]:
        srv_entry = self.defs.get(service_name)
        if srv_entry is not None:
            return srv_entry[1]
        return None

    def sort(self):
        self.defs = {key: self.defs[key] for key in sorted(self.defs.keys())}



from datetime import timedelta
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock

from spacepackets.ecss import PusTelecommand
from tmtccmd import CcsdsTmtcBackend, CcsdsTmListener, TcHandlerBase
from tmtccmd.com_if import ComInterface
from tmtccmd.com_if.dummy import DummyComIF
from tmtccmd.core import TcMode, TmMode, BackendRequest
from tmtccmd.core.ccsds_backend import NoValidProcedureSet
from tmtccmd.tc import (
    TcProcedureBase,
    DefaultProcedureInfo,
    TcProcedureType,
    ProcedureHelper,
    TcQueueEntryBase,
)
from tmtccmd.tc.handler import FeedWrapper


class TcHandlerMock(TcHandlerBase):
    def __init__(self):
        super().__init__()
        self.is_feed_cb_valid = False
        self.feed_cb_call_count = 0
        self.feed_cb_def_proc_count = 0
        self.send_cb_call_count = 0
        self.send_cb_call_args = None
        self.send_cb_service_arg: Optional[str] = None
        self.send_cb_op_code_arg: Optional[str] = None

    def send_cb(self, entry_helper: ProcedureHelper, com_if: ComInterface):
        self.send_cb_call_count += 1
        self.send_cb_call_args = (entry_helper, com_if)

    def queue_finished_cb(self, info: TcProcedureBase):
        pass

    def feed_cb(self, info: ProcedureHelper, wrapper: FeedWrapper):
        self.feed_cb_call_count += 1
        self.send_cb_service_arg = None
        self.send_cb_op_code_arg = None
        if info is not None:
            if info.proc_type == TcProcedureType.DEFAULT:
                self.feed_cb_def_proc_count += 1
                def_info = info.to_def_procedure()
                if def_info.service != "17":
                    self.is_feed_cb_valid = False
                self.send_cb_service_arg = def_info.service
                self.send_cb_op_code_arg = def_info.op_code
                if def_info.service == "17":
                    if def_info.op_code == "0":
                        wrapper.queue_helper.add_pus_tc(
                            PusTelecommand(service=17, subservice=1)
                        )
                    elif def_info.op_code == "1":
                        wrapper.queue_helper.add_pus_tc(
                            PusTelecommand(service=17, subservice=1)
                        )
                        wrapper.queue_helper.add_pus_tc(
                            PusTelecommand(service=5, subservice=1)
                        )
                    elif def_info.op_code == "2":
                        wrapper.queue_helper.add_pus_tc(
                            PusTelecommand(service=17, subservice=1)
                        )
                        wrapper.queue_helper.add_wait(timedelta(milliseconds=20))
                        wrapper.queue_helper.add_pus_tc(
                            PusTelecommand(service=5, subservice=1)
                        )


class TestBackend(TestCase):
    def setUp(self) -> None:
        self.com_if = DummyComIF()
        self.tm_listener = MagicMock(specs=CcsdsTmListener)
        self.tc_handler = TcHandlerMock()
        self.backend = CcsdsTmtcBackend(
            tc_mode=TcMode.IDLE,
            tm_mode=TmMode.IDLE,
            com_if=self.com_if,
            tm_listener=self.tm_listener,
            tc_handler=self.tc_handler,
        )
        self.assertEqual(self.backend.tm_listener, self.tm_listener)

    def test_idle(self):
        self.assertEqual(self.backend.tm_mode, TmMode.IDLE)
        self.assertEqual(self.backend.tc_mode, TcMode.IDLE)
        self.assertEqual(self.backend.com_if.get_id(), "dummy")
        self.assertEqual(self.backend.com_if_id, "dummy")
        self.assertEqual(self.backend.inter_cmd_delay, timedelta())
        state = self.backend.state
        self.assertEqual(state.tc_mode, TcMode.IDLE)
        self.assertEqual(state.tm_mode, TmMode.IDLE)
        self.assertEqual(state.next_delay, timedelta())
        self.assertEqual(state.request, BackendRequest.NONE)
        self.backend.mode_to_req()
        self.assertEqual(state.request, BackendRequest.DELAY_IDLE)
        self.assertEqual(self.backend.inter_cmd_delay, timedelta())
        self.assertFalse(self.backend.com_if_active())
        self.assertFalse(self.com_if.is_open())

    def test_basic_ops(self):
        self.backend.start()
        self.assertTrue(self.com_if.is_open())
        self.assertTrue(self.backend.com_if_active())

        res = self.backend.periodic_op()
        self.assertEqual(res.request, BackendRequest.DELAY_IDLE)
        self.backend.tm_mode = TmMode.LISTENER
        self.assertEqual(self.backend.tm_mode, TmMode.LISTENER)
        res = self.backend.periodic_op()
        self.tm_listener.operation.assert_called_once()
        self.backend.poll_tm()
        self.assertEqual(self.tm_listener.operation.call_count, 2)
        self.assertEqual(res.request, BackendRequest.DELAY_LISTENER)
        self.backend.tm_mode = TmMode.IDLE
        self.backend.tc_mode = TcMode.ONE_QUEUE
        with self.assertRaises(NoValidProcedureSet):
            self.backend.periodic_op()
        self.backend.current_procedure = DefaultProcedureInfo(service="17", op_code="0")

        res = self.backend.periodic_op()
        # Only one queue entry which is handled immediately
        self.assertEqual(res.request, BackendRequest.TERMINATION_NO_ERROR)
        self.assertEqual(self.tc_handler.feed_cb_def_proc_count, 1)
        self.assertEqual(self.tc_handler.feed_cb_call_count, 1)
        self.assertEqual(self.tc_handler.send_cb_call_count, 1)
        self.assertIsNotNone(self.tc_handler.send_cb_call_args)
        self.assertIsNotNone(self.tc_handler.send_cb_call_args[0])
        self.assertIsNotNone(self.tc_handler.send_cb_call_args[1])
        self.assertEqual(self.tc_handler.send_cb_call_args[1], self.com_if)
        cast_wrapper = self.tc_handler.send_cb_call_args[0]
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(pus_entry.pus_tc, PusTelecommand(service=17, subservice=1))
        self.backend.close_com_if()
        self.assertFalse(self.com_if.is_open())

    def test_one_queue_multi_entry_ops(self):
        self.backend.tm_mode = TmMode.IDLE
        self.backend.tc_mode = TcMode.ONE_QUEUE
        self.backend.current_procedure = DefaultProcedureInfo(service="17", op_code="1")
        res = self.backend.periodic_op()
        self.assertEqual(res.request, BackendRequest.CALL_NEXT)
        self.assertEqual(self.tc_handler.feed_cb_def_proc_count, 1)
        self.assertEqual(self.tc_handler.feed_cb_call_count, 1)
        self.assertEqual(self.tc_handler.send_cb_call_count, 1)
        self._check_tc_req_recvd(17, 1)
        res = self.backend.periodic_op()
        self.assertEqual(self.tc_handler.feed_cb_def_proc_count, 1)
        self.assertEqual(self.tc_handler.feed_cb_call_count, 1)
        self.assertEqual(self.tc_handler.send_cb_call_count, 2)
        self._check_tc_req_recvd(5, 1)
        self.assertEqual(res.request, BackendRequest.TERMINATION_NO_ERROR)

    def test_multi_queue_ops(self):
        self.backend.tm_mode = TmMode.IDLE
        self.backend.tc_mode = TcMode.MULTI_QUEUE
        self.backend.current_procedure = DefaultProcedureInfo(service="17", op_code="0")
        res = self.backend.periodic_op()
        self.assertEqual(res.request, BackendRequest.CALL_NEXT)
        self.assertEqual(self.backend.request, BackendRequest.CALL_NEXT)
        self.assertEqual(self.backend.tc_mode, TcMode.IDLE)
        self.assertEqual(self.tc_handler.feed_cb_def_proc_count, 1)
        self.assertEqual(self.tc_handler.feed_cb_call_count, 1)
        self.assertEqual(self.tc_handler.send_cb_call_count, 1)
        self._check_tc_req_recvd(17, 1)
        res = self.backend.periodic_op()
        self.assertEqual(self.tc_handler.feed_cb_call_count, 1)
        self.assertEqual(res.request, BackendRequest.DELAY_IDLE)
        self.backend.tc_mode = TcMode.MULTI_QUEUE
        self.backend.current_procedure = DefaultProcedureInfo(service="17", op_code="0")
        res = self.backend.periodic_op()
        self.assertEqual(res.request, BackendRequest.CALL_NEXT)
        self.assertEqual(self.backend.request, BackendRequest.CALL_NEXT)
        self.assertEqual(self.tc_handler.feed_cb_def_proc_count, 2)
        self.assertEqual(self.tc_handler.feed_cb_call_count, 2)

    def test_procedure_handling(self):
        def_proc = DefaultProcedureInfo(service="17", op_code="0")
        self.backend.current_procedure = def_proc
        self.assertEqual(
            self.backend.current_procedure.proc_type, TcProcedureType.DEFAULT
        )
        proc_helper = self.backend.current_procedure
        def_proc = proc_helper.to_def_procedure()
        self.assertIsNotNone(def_proc)
        self.assertEqual(def_proc.service, "17")
        self.assertEqual(def_proc.op_code, "0")

    def _check_tc_req_recvd(self, service: int, subservice: int):
        self.assertEqual(self.tc_handler.send_cb_call_args[1], self.com_if)
        cast_wrapper = self.tc_handler.send_cb_call_args[0]
        pus_entry = cast_wrapper.to_pus_tc_entry()
        self.assertEqual(
            pus_entry.pus_tc, PusTelecommand(service=service, subservice=subservice)
        )



"""PyQt front end components for the tmtccmd framework.
@author         R. Mueller, P. Scheurenbrand, D. Nguyen
"""
import os
import sys
import webbrowser
from multiprocessing import Process
from pathlib import Path
from typing import Union

from PyQt5.QtWidgets import (
    QMainWindow,
    QGridLayout,
    QTableWidget,
    QWidget,
    QLabel,
    QCheckBox,
    QDoubleSpinBox,
    QFrame,
    QComboBox,
    QPushButton,
    QTableWidgetItem,
    QMenu,
    QAction,
)
from PyQt5.QtGui import QPixmap, QIcon, QFont
from PyQt5.QtCore import (
    Qt,
    QThreadPool,
)

from tmtccmd.core.base import FrontendBase
from tmtccmd.config.globals import CoreGlobalIds
from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend
from tmtccmd.config import TmTcCfgHookBase
from tmtccmd.gui.buttons import (
    ConnectButtonParams,
    ButtonArgs,
    SendButtonWrapper,
    TmButtonWrapper,
    ConnectButtonWrapper,
)
from tmtccmd.gui.defs import SharedArgs, CONNECT_BTTN_STYLE, FrontendState
from tmtccmd.logging import get_console_logger
from tmtccmd.core.globals_manager import get_global, update_global
from tmtccmd.com_if.tcpip_utils import TcpIpConfigIds
import tmtccmd as mod_root

LOGGER = get_console_logger()


class TmTcFrontend(QMainWindow, FrontendBase):
    def __init__(
        self, hook_obj: TmTcCfgHookBase, tmtc_backend: CcsdsTmtcBackend, app_name: str
    ):
        super(TmTcFrontend, self).__init__()
        super(QMainWindow, self).__init__()
        self._app_name = app_name
        self._shared_args = SharedArgs(tmtc_backend)
        tmtc_backend.exit_on_com_if_init_failure = False
        self._hook_obj = hook_obj
        self._service_list = []
        self._op_code_list = []
        self._com_if_list = []
        self._service_op_code_dict = hook_obj.get_tmtc_definitions()
        self._state = FrontendState()
        self._thread_pool = QThreadPool()
        self.__connected = False
        self.__debug_mode = True

        self.__combo_box_op_codes: Union[None, QComboBox] = None
        self.logo_path = Path(
            f"{Path(mod_root.__file__).parent.parent}/misc/logo-tiny.png"
        )

    def prepare_start(self, args: any) -> Process:
        return Process(target=self.start)

    def start(self, qt_app: any):
        self.__start_ui()
        sys.exit(qt_app.exec())

    def set_gui_logo(self, logo_total_path: str):
        if os.path.isfile(logo_total_path):
            self.logo_path = logo_total_path
        else:
            LOGGER.warning("Could not set logo, path invalid!")

    def __start_ui(self):
        self.__create_menu_bar()
        win = QWidget(self)
        self.setCentralWidget(win)

        grid = QGridLayout()
        win.setLayout(grid)
        row = 0
        self.setWindowTitle(self._app_name)
        print(self.logo_path)
        self.setWindowIcon(QIcon(self.logo_path.as_posix()))

        add_pixmap = False

        if add_pixmap:
            row = self.__set_up_pixmap(grid=grid, row=row)

        row = self.__set_up_config_section(grid=grid, row=row)
        row = self.__add_vertical_separator(grid=grid, row=row)

        tm_listener_button = QPushButton()
        conn_bttn_params = ConnectButtonParams(
            hook_obj=self._hook_obj,
            connect_cb=self.__connected_com_if_cb,
            disconnect_cb=self.__disconnect_com_if_cb,
            tm_listener_bttn=tm_listener_button,
        )
        # com if configuration
        row, self.__connect_button_wrapper = self.__set_up_com_if_section(
            conn_bttn_params=conn_bttn_params, grid=grid, row=row
        )
        row = self.__add_vertical_separator(grid=grid, row=row)

        tmtc_ctrl_label = QLabel("TMTC Control")
        font = QFont()
        font.setBold(True)
        tmtc_ctrl_label.setFont(font)
        grid.addWidget(tmtc_ctrl_label, row, 0, 1, 2)
        row += 1
        row = self.__set_up_service_op_code_section(grid=grid, row=row)

        button_args = ButtonArgs(
            state=self._state, pool=self._thread_pool, shared=self._shared_args
        )
        self.__send_bttn_wrapper = SendButtonWrapper(
            button=QPushButton(),
            args=button_args,
            conn_button=self.__connect_button_wrapper.button,
        )
        grid.addWidget(self.__send_bttn_wrapper.button, row, 0, 1, 2)
        row += 1

        self.__tm_button_wrapper = TmButtonWrapper(
            button=tm_listener_button,
            args=button_args,
            conn_button=self.__connect_button_wrapper.button,
        )
        grid.addWidget(self.__tm_button_wrapper.button, row, 0, 1, 2)
        row += 1
        self.show()

    def __create_menu_bar(self):
        menu_bar = self.menuBar()
        # Creating menus using a QMenu object
        file_menu = QMenu("&File", self)
        menu_bar.addMenu(file_menu)
        # Creating menus using a title
        help_menu = menu_bar.addMenu("&Help")

        help_action = QAction("Help", self)
        help_action.triggered.connect(self.__help_url)
        help_menu.addAction(help_action)

    @staticmethod
    def __help_url():
        webbrowser.open("https://tmtccmd.readthedocs.io/en/latest/")

    def __set_up_config_section(self, grid: QGridLayout, row: int) -> int:
        label = QLabel("Configuration")
        font = QFont()
        font.setBold(True)
        label.setFont(font)
        grid.addWidget(label, row, 0, 1, 2)
        row += 1

        start_listener_on_connect = QCheckBox("Auto-Connect TM listener")
        start_listener_on_connect.setChecked(True)
        start_listener_on_connect.stateChanged.connect(
            lambda: self._tm_auto_connect_changed(start_listener_on_connect)
        )
        grid.addWidget(start_listener_on_connect, row, 0, 1, 1)
        row += 1

        grid.addWidget(QLabel("Inter-Packet Delay Seconds [0 - 500]"), row, 0, 1, 2)
        row += 1

        spin_timeout = QDoubleSpinBox()
        spin_timeout.setValue(0.1)
        # TODO: set sensible min/max values
        spin_timeout.setSingleStep(0.1)
        spin_timeout.setMinimum(0.0)
        spin_timeout.setMaximum(500.0)
        # https://youtrack.jetbrains.com/issue/PY-22908
        # Ignore those warnings for now.
        spin_timeout.valueChanged.connect(number_timeout)
        grid.addWidget(spin_timeout, row, 0, 1, 1)
        row += 1
        return row

    def _tm_auto_connect_changed(self, box: QCheckBox):
        if box.isChecked():
            self._state.auto_connect_tm_listener = True
        else:
            self._state.auto_connect_tm_listener = False

    def __set_up_com_if_section(
        self, conn_bttn_params: ConnectButtonParams, grid: QGridLayout, row: int
    ) -> (int, ConnectButtonWrapper):
        font = QFont()
        font.setBold(True)
        label = QLabel("Communication Interface")
        label.setFont(font)
        grid.addWidget(label, row, 0, 1, 1)
        com_if_combo_box = QComboBox()
        all_com_ifs = self._hook_obj.get_com_if_dict()
        index = 0
        # add all possible ComIFs to the comboBox
        for id, com_if_value in all_com_ifs.items():
            com_if_combo_box.addItem(com_if_value[0])
            self._com_if_list.append((id, com_if_value[0]))
            if self._shared_args.backend.com_if_id == id:
                com_if_combo_box.setCurrentIndex(index)
            index += 1
        com_if_combo_box.currentIndexChanged.connect(self.__com_if_sel_index_changed)
        grid.addWidget(com_if_combo_box, row, 1, 1, 1)
        row += 1

        self.com_if_cfg_button = QPushButton()
        self.com_if_cfg_button.setText("Configure")
        grid.addWidget(self.com_if_cfg_button, row, 0, 1, 2)
        row += 1

        connect_button = QPushButton()
        connect_button.setText("Connect")
        connect_button.setStyleSheet(CONNECT_BTTN_STYLE)
        conn_bttn_wrapper = ConnectButtonWrapper(
            button=connect_button,
            args=ButtonArgs(self._state, self._thread_pool, self._shared_args),
            bttn_params=conn_bttn_params,
        )
        grid.addWidget(connect_button, row, 0, 1, 2)
        row += 1
        return row, conn_bttn_wrapper

    def __disable_conn_bttn(self):
        self.__connect_button_wrapper.button.setDisabled(True)

    def __enable_conn_bttn(self):
        self.__connect_button_wrapper.button.setEnabled(True)

    def __connected_com_if_cb(self):
        self.__send_bttn_wrapper.button.setEnabled(True)
        self.__tm_button_wrapper.button.setEnabled(True)

    def __disconnect_com_if_cb(self):
        self.__send_bttn_wrapper.button.setDisabled(True)
        self.__tm_button_wrapper.button.setDisabled(True)

    def __set_up_service_op_code_section(self, grid: QGridLayout, row: int):
        grid.addWidget(QLabel("Service: "), row, 0, 1, 2)
        grid.addWidget(QLabel("Operation Code: "), row, 1, 1, 2)
        row += 1

        combo_box_services = QComboBox()
        default_service = get_global(CoreGlobalIds.CURRENT_SERVICE)
        self._service_op_code_dict = self._hook_obj.get_tmtc_definitions()
        if self._service_op_code_dict is None:
            LOGGER.warning("Invalid service to operation code dictionary")
            LOGGER.warning("Setting default dictionary")
            from tmtccmd.config.globals import get_default_tmtc_defs

            self._service_op_code_dict = get_default_tmtc_defs()
        index = 0
        default_index = 0
        for service_key, service_value in self._service_op_code_dict.defs.items():
            combo_box_services.addItem(service_value[0])
            if service_key == default_service:
                default_index = index
            self._service_list.append(service_key)
            index += 1
        combo_box_services.setCurrentIndex(default_index)
        self._state.current_service = self._service_list[default_index]

        combo_box_services.currentIndexChanged.connect(self.__service_index_changed)
        grid.addWidget(combo_box_services, row, 0, 1, 1)

        self.__combo_box_op_codes = QComboBox()
        self._state.current_service = self._service_list[default_index]
        self.__update_op_code_combo_box()
        self.__combo_box_op_codes.currentIndexChanged.connect(
            self.__op_code_index_changed
        )
        # TODO: Combo box also needs to be updated if another service is selected
        grid.addWidget(self.__combo_box_op_codes, row, 1, 1, 1)
        row += 1
        return row

    def __set_up_pixmap(self, grid: QGridLayout, row: int) -> int:
        label = QLabel(self)
        label.setGeometry(720, 10, 100, 100)
        label.adjustSize()

        pixmap = QPixmap(self.logo_path)
        pixmap_width = pixmap.width()
        pixmap_height = pixmap.height()
        row += 1

        pixmap_scaled = pixmap.scaled(
            pixmap_width * 0.3, pixmap_height * 0.3, Qt.KeepAspectRatio
        )
        label.setPixmap(pixmap_scaled)
        label.setScaledContents(True)

        grid.addWidget(label, row, 0, 1, 2)
        row += 1
        return row

    @staticmethod
    def __add_vertical_separator(grid: QGridLayout, row: int):
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        grid.addWidget(separator, row, 0, 1, 2)
        row += 1
        return row

    def __service_index_changed(self, index: int):
        self._current_service = self._service_list[index]
        self.__update_op_code_combo_box()
        if self.__debug_mode:
            LOGGER.info("Service changed")

    def __op_code_index_changed(self, index: int):
        self._state.current_op_code = self._op_code_list[index]
        if self.__debug_mode:
            LOGGER.info("Op Code changed")

    def __update_op_code_combo_box(self):
        self.__combo_box_op_codes.clear()
        self._op_code_list = []
        op_code_entry = self._service_op_code_dict.op_code_entry(
            self._state.current_service
        )
        if op_code_entry is not None:
            for op_code_key, op_code_value in op_code_entry.op_code_dict.items():
                try:
                    self._op_code_list.append(op_code_key)
                    self.__combo_box_op_codes.addItem(op_code_value[0])
                except TypeError:
                    LOGGER.warning(f"Invalid op code entry {op_code_value}, skipping..")
            self._state.current_op_code = self._op_code_list[0]

    def __checkbox_log_update(self, state: int):
        update_global(CoreGlobalIds.PRINT_TO_FILE, state)
        if self.__debug_mode:
            LOGGER.info(["Enabled", "Disabled"][state == 0] + " print to log.")

    def __checkbox_console_update(self, state: bool):
        update_global(CoreGlobalIds.PRINT_TM, state)
        if self.__debug_mode:
            LOGGER.info(["enabled", "disabled"][state == 0] + " console print")

    def __checkbox_print_raw_data_update(self, state: int):
        update_global(CoreGlobalIds.PRINT_RAW_TM, state)
        if self.__debug_mode:
            LOGGER.info(["enabled", "disabled"][state == 0] + " printing of raw data")

    def __com_if_sel_index_changed(self, index: int):
        self._state.current_com_if = self._com_if_list[index][0]
        if self.__debug_mode:
            LOGGER.info(f"Communication IF updated: {self._com_if_list[index][1]}")


class SingleCommandTable(QTableWidget):
    def __init__(self):
        super().__init__()
        self.setRowCount(1)
        self.setColumnCount(5)
        self.setHorizontalHeaderItem(0, QTableWidgetItem("Service"))
        self.setHorizontalHeaderItem(1, QTableWidgetItem("Subservice"))
        self.setHorizontalHeaderItem(2, QTableWidgetItem("SSC"))
        self.setHorizontalHeaderItem(3, QTableWidgetItem("Data"))
        self.setHorizontalHeaderItem(4, QTableWidgetItem("CRC"))
        self.setItem(0, 0, QTableWidgetItem("17"))
        self.setItem(0, 1, QTableWidgetItem("1"))
        self.setItem(0, 2, QTableWidgetItem("20"))


def checkbox_print_hk_data(state: int):
    update_global(CoreGlobalIds.PRINT_HK, state)
    LOGGER.info(["enabled", "disabled"][state == 0] + " printing of hk data")


def checkbox_short_display_mode(state: int):
    update_global(CoreGlobalIds.DISPLAY_MODE, state)
    LOGGER.info(["enabled", "disabled"][state == 0] + " short display mode")


def number_timeout(value: float):
    update_global(CoreGlobalIds.TM_TIMEOUT, value)
    LOGGER.info("PUS TM timeout changed to: " + str(value))


def number_timeout_factor(value: float):
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, value)
    LOGGER.info("PUS TM timeout factor changed to: " + str(value))


def ip_change_client(value):
    ethernet_config = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    ethernet_config[TcpIpConfigIds.RECV_ADDRESS] = value
    update_global(CoreGlobalIds.ETHERNET_CONFIG, ethernet_config)
    LOGGER.info("Client IP changed: " + value)


def ip_change_board(value):
    ethernet_config = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    ethernet_config[TcpIpConfigIds.SEND_ADDRESS] = value
    update_global(CoreGlobalIds.ETHERNET_CONFIG, ethernet_config)
    LOGGER.info("Board IP changed: " + value)
